@@ -35,7 +35,7 @@
 #
 # Wind and solar capacities and locations: EEG Stammdaten, based on  http://www.energymap.info/download.html, which represents capacities at the end of 2014. Units without PLZ are removed.
 #
-# Wind and solar time series: REatlas, Andresen et al, "Validation of Danish wind time series from a new global renewable energy atlas for energy system analysis," Energy 93 (2015) 1074 - 1088.
+# Wind and solar time series: REatlas, Andresen et al, "Validation of Danish wind time series from a new global renewable energy atlas for energy system analysis, " Energy 93 (2015) 1074 - 1088.
 #
 # NB:
 #
@@ -73,13 +73,12 @@
 
 import os
 
-import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 
 import pypsa
 
-#%matplotlib inline
+# %matplotlib inline
 
 ### Read in the raw SciGRID data
 
@@ -225,7 +224,6 @@ start = 0  # km
 stop = 1  # km
 
 for (u, v), dist in pairs.sort_values().iteritems():
-
     if dist < start:
         continue
 
@@ -257,7 +255,6 @@ for (u, v), dist in pairs.sort_values().iteritems():
 limit = 0.85
 
 for (u, v), dist in pairs.sort_values().iteritems():
-
     # only go up to pairs stop km apart
     if dist > limit:
         break
@@ -361,7 +358,7 @@ for sn in network.sub_networks.obj:
             )
         )
         # print(buses.index)
-        # print(len(branches),branches.index)
+        # print(len(branches), branches.index)
         for bus in buses:
             network.remove("Bus", bus)
         for branch in branches:
@@ -439,7 +436,6 @@ def generate_dummy_graph(network):
     when the buses have been split, so that all load and generation is
     attached to the 220kV bus.
     """
-
     graph = pypsa.descriptors.OrderedGraph()
 
     graph.add_nodes_from(
@@ -508,7 +504,7 @@ for bus in graph.nodes():
         ),
     )
 
-#%matplotlib inline
+# %matplotlib inline
 
 pd.DataFrame(load.sum(axis=1)).plot()
 
@@ -528,7 +524,6 @@ monthly_load.plot(grid=True)
 
 
 def read_kraftwerksliste(with_latlon=True):
-
     kraftwerke = pd.read_csv(
         "../../lib/vresutils/data/Kraftwerksliste_CSV_deCP850ed.csv",
         delimiter=";",
@@ -564,7 +559,7 @@ def read_kraftwerksliste(with_latlon=True):
         }
         kraftwerke["lon"] = kraftwerke.PLZ.map({pc: c.x for pc, c in postcodes.items()})
         kraftwerke["lat"] = kraftwerke.PLZ.map({pc: c.y for pc, c in postcodes.items()})
-        # kraftwerke.dropna(subset=('lon','lat'), inplace=True)
+        # kraftwerke.dropna(subset=('lon', 'lat'), inplace=True)
 
     kraftwerke["Type"] = kraftwerke["Auswertung Energieträger"].map(
         {
@@ -608,7 +603,6 @@ import random
 
 
 def backup_capacity_german_grid(G):
-
     from shapely.geometry import Point
 
     plants = power_plants
@@ -669,7 +663,7 @@ m_costs = {
 
 default_cost = 10.0
 
-for (bus_name, tech_name) in cap.index:
+for bus_name, tech_name in cap.index:
     print(bus_name, tech_name, cap[(bus_name, tech_name)])
     if tech_name == "Pumped Hydro":
         network.add(
