@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-
 """
 Power system components.
 """
@@ -11,7 +10,7 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
@@ -22,6 +21,7 @@ from pathlib import Path
 
 import numpy as np
 import pandas as pd
+import validators
 from scipy.sparse import csgraph
 
 from pypsa.contingency import calculate_BODF, network_lpf_contingency, network_sclopf
@@ -137,7 +137,7 @@ class Network(Basic):
     ----------
     import_name : string, Path
         Path to netCDF file, HDF5 .h5 store or folder of CSV files from which to
-        import network data.
+        import network data. The string could be a URL.
     name : string, default ""
         Network name.
     ignore_standard_types : boolean, default False
@@ -162,6 +162,7 @@ class Network(Basic):
     --------
     >>> nw1 = pypsa.Network("my_store.h5")
     >>> nw2 = pypsa.Network("/my/folder")
+    >>> nw3 = pypsa.Network("https://github.com/PyPSA/PyPSA/raw/master/examples/scigrid-de/scigrid-with-load-gen-trafos.nc")
     """
 
     # Spatial Reference System Identifier (SRID), defaults to longitude and latitude
@@ -232,7 +233,6 @@ class Network(Basic):
         override_component_attrs=None,
         **kwargs,
     ):
-
         # Initialise root logger and set its level, if this has not been done before
         logging.basicConfig(level=logging.INFO)
 
@@ -333,10 +333,11 @@ class Network(Basic):
             self.read_in_default_standard_types()
 
         if import_name:
-            import_name = Path(import_name)
-            if import_name.suffix == ".h5":
+            if not validators.url(str(import_name)):
+                import_name = Path(import_name)
+            if str(import_name).endswith(".h5"):
                 self.import_from_hdf5(import_name)
-            elif import_name.suffix == ".nc":
+            elif str(import_name).endswith(".nc"):
                 self.import_from_netcdf(import_name)
             elif import_name.is_dir():
                 self.import_from_csv_folder(import_name)
@@ -369,9 +370,7 @@ class Network(Basic):
         Function called when network is created to build component
         pandas.DataFrames.
         """
-
         for component in self.all_components:
-
             attrs = self.components[component]["attrs"]
 
             static_dtypes = attrs.loc[attrs.static, "dtype"].drop(["name"])
@@ -396,9 +395,7 @@ class Network(Basic):
             setattr(self, self.components[component]["list_name"] + "_t", pnl)
 
     def read_in_default_standard_types(self):
-
         for std_type in self.standard_type_components:
-
             list_name = self.components[std_type]["list_name"]
 
             file_name = os.path.join(
@@ -639,6 +636,7 @@ class Network(Basic):
         solver_options={},
         solver_logfile=None,
         formulation="kirchhoff",
+        transmission_losses=0,
         keep_files=False,
         extra_functionality=None,
         multi_investment_periods=False,
@@ -668,10 +666,15 @@ class Network(Basic):
             construction, e.g. .lp file - useful for debugging
         formulation : string
             Formulation of the linear power flow equations to use; must be
-            one of ["angles","cycles","kirchhoff","ptdf"]
+            one of ["angles", "cycles", "kirchhoff", "ptdf"]
+        transmission_losses : int
+            Whether an approximation of transmission losses should be included
+            in the linearised power flow formulation. A passed number will denote
+            the number of tangents used for the piecewise linear approximation.
+            Defaults to 0, which ignores losses.
         extra_functionality : callable function
             This function must take two arguments
-            `extra_functionality(network,snapshots)` and is called after
+            `extra_functionality(network, snapshots)` and is called after
             the model building is complete, but before it is sent to the
             solver. It allows the user to
             add/change constraints and add/change the objective function.
@@ -700,7 +703,7 @@ class Network(Basic):
         extra_postprocessing : callable function
             Only taking effect when pyomo is True.
             This function must take three arguments
-            `extra_postprocessing(network,snapshots,duals)` and is called after
+            `extra_postprocessing(network, snapshots, duals)` and is called after
             the model has solved and the results are extracted. It allows the user
             to extract further information about the solution, such as additional
             shadow prices.
@@ -750,6 +753,7 @@ class Network(Basic):
             "keep_files": keep_files,
             "solver_options": solver_options,
             "formulation": formulation,
+            "transmission_losses": transmission_losses,
             "extra_functionality": extra_functionality,
             "multi_investment_periods": multi_investment_periods,
             "solver_name": solver_name,
@@ -766,8 +770,7 @@ class Network(Basic):
 
         if pyomo:
             return network_lopf(self, **args)
-        else:
-            return network_lopf_lowmem(self, **args)
+        return network_lopf_lowmem(self, **args)
 
     def add(self, class_name, name, **kwargs):
         """
@@ -793,11 +796,10 @@ class Network(Basic):
 
         Examples
         --------
-        >>> network.add("Bus","my_bus_0")
-        >>> network.add("Bus","my_bus_1",v_nom=380)
-        >>> network.add("Line","my_line_name",bus0="my_bus_0",bus1="my_bus_1",length=34,r=2,x=4)
+        >>> network.add("Bus", "my_bus_0")
+        >>> network.add("Bus", "my_bus_1", v_nom=380)
+        >>> network.add("Line", "my_line_name", bus0="my_bus_0", bus1="my_bus_1", length=34, r=2, x=4)
         """
-
         assert class_name in self.components, "Component class {} not found".format(
             class_name
         )
@@ -866,9 +868,8 @@ class Network(Basic):
 
         Examples
         --------
-        >>> network.remove("Line","my_line 12345")
+        >>> network.remove("Line", "my_line 12345")
         """
-
         if class_name not in self.components:
             logger.error("Component class {} not found".format(class_name))
             return None
@@ -906,7 +907,7 @@ class Network(Basic):
             All components are named after names with this added suffix. It
             is assumed that all Series and DataFrames are indexed by the original names.
         kwargs
-            Component attributes, e.g. x=[0.1,0.2], can be list, pandas.Series
+            Component attributes, e.g. x=[0.1, 0.2], can be list, pandas.Series
             of pandas.DataFrame for time-varying
 
         Returns
@@ -920,8 +921,8 @@ class Network(Basic):
         Short Example:
 
         >>> network.madd("Load", ["load 1", "load 2"],
-        ...        bus=["1","2"],
-        ...        p_set=np.random.rand(len(network.snapshots),2))
+        ...        bus=["1", "2"],
+        ...        p_set=np.random.rand(len(network.snapshots), 2))
 
         Long Example:
 
@@ -935,9 +936,9 @@ class Network(Basic):
         >>> n.madd("Load",
         ...        n.buses.index + " load",
         ...        bus=buses,
-        ...        p_set=np.random.rand(len(snapshots),len(buses)))
+        ...        p_set=np.random.rand(len(snapshots), len(buses)))
         >>> # add wind availability as pandas DataFrame
-        >>> wind = pd.DataFrame(np.random.rand(len(snapshots),len(buses)),
+        >>> wind = pd.DataFrame(np.random.rand(len(snapshots), len(buses)),
         ...        index=n.snapshots,
         ...        columns=buses)
         >>> #use a suffix to avoid boilerplate to rename everything
@@ -949,7 +950,6 @@ class Network(Basic):
         ...        capital_cost=1e5,
         ...        p_max_pu=wind)
         """
-
         if class_name not in self.components:
             logger.error("Component class {} not found".format(class_name))
             return None
@@ -1000,7 +1000,6 @@ class Network(Basic):
         --------
         >>> network.mremove("Line", ["line x", "line y"])
         """
-
         if class_name not in self.components:
             logger.error("Component class {} not found".format(class_name))
             return None
@@ -1018,7 +1017,6 @@ class Network(Basic):
             df.drop(df.columns.intersection(names), axis=1, inplace=True)
 
     def _retrieve_overridden_components(self):
-
         components_index = list(self.components.keys())
 
         cols = ["list_name", "description", "type"]
@@ -1136,7 +1134,6 @@ class Network(Basic):
 
         >>> sub_network_0_with_only_10_snapshots = network[:10, network.buses.sub_network = "0"]
         """
-
         if isinstance(key, tuple):
             time_i, key = key
         else:
@@ -1222,7 +1219,9 @@ class Network(Basic):
             sort=True,
         )
 
-    def determine_network_topology(self, investment_period=None):
+    def determine_network_topology(
+        self, investment_period=None, skip_isolated_buses=False
+    ):
         """
         Build sub_networks from topology.
 
@@ -1232,7 +1231,6 @@ class Network(Basic):
         the network topology is determined on the basis of the active
         branches.
         """
-
         adjacency_matrix = self.adjacency_matrix(
             branch_components=self.passive_branch_components,
             investment_period=investment_period,
@@ -1250,6 +1248,10 @@ class Network(Basic):
         for i in np.arange(n_components):
             # index of first bus
             buses_i = (labels == i).nonzero()[0]
+
+            if skip_isolated_buses and (len(buses_i) == 1):
+                continue
+
             carrier = self.buses.carrier.iat[buses_i[0]]
 
             if carrier not in ["AC", "DC"] and len(buses_i) > 1:
@@ -1304,7 +1306,7 @@ class Network(Basic):
             if not (skip_empty and self.df(c).empty)
         )
 
-    def consistency_check(self):
+    def consistency_check(self, check_dtypes=False):
         """
         Checks the network for consistency; e.g. that all components are
         connected to existing buses and that no impedances are singular.
@@ -1327,6 +1329,10 @@ class Network(Basic):
         for c in self.iterate_components():
             for attr in bus_columns(c.df):
                 missing = ~c.df[attr].isin(self.buses.index)
+                # if bus2, bus3... contain empty strings do not warn
+                if c.name in self.branch_components:
+                    if int(attr[-1]) > 1:
+                        missing &= c.df[attr] != ""
                 if missing.any():
                     msg = "The following %s have %s which are not defined:\n%s"
                     logger.warning(msg, c.list_name, attr, c.df.index[missing])
@@ -1487,48 +1493,61 @@ class Network(Basic):
                         f"\n\n\t{', '.join(intersection)}"
                     )
 
+            if c.name in {"Generator"}:
+                bad_uc_gens = c.df.index[
+                    c.df.committable
+                    & (c.df.up_time_before > 0)
+                    & (c.df.down_time_before > 0)
+                ]
+                if not bad_uc_gens.empty:
+                    logger.warning(
+                        "The following committable generators were both up and down"
+                        f" before the simulation: {bad_uc_gens}."
+                        " This could cause an infeasibility."
+                    )
+
         # check all dtypes of component attributes
 
-        for c in self.iterate_components():
+        if check_dtypes:
+            for c in self.iterate_components():
+                # first check static attributes
 
-            # first check static attributes
-
-            dtypes_soll = c.attrs.loc[c.attrs["static"], "dtype"].drop("name")
-            unmatched = c.df.dtypes[dtypes_soll.index] != dtypes_soll
-
-            if unmatched.any():
-                logger.warning(
-                    "The following attributes of the dataframe %s "
-                    "have the wrong dtype:\n%s\n"
-                    "They are:\n%s\nbut should be:\n%s",
-                    c.list_name,
-                    unmatched.index[unmatched],
-                    c.df.dtypes[dtypes_soll.index[unmatched]],
-                    dtypes_soll[unmatched],
-                )
-
-            # now check varying attributes
-
-            types_soll = c.attrs.loc[c.attrs["varying"], ["typ", "dtype"]]
-
-            for attr, typ, dtype in types_soll.itertuples():
-                if c.pnl[attr].empty:
-                    continue
-
-                unmatched = c.pnl[attr].dtypes != dtype
+                dtypes_soll = c.attrs.loc[c.attrs["static"], "dtype"].drop("name")
+                unmatched = c.df.dtypes[dtypes_soll.index] != dtypes_soll
 
                 if unmatched.any():
                     logger.warning(
-                        "The following columns of time-varying attribute "
-                        "%s in %s_t have the wrong dtype:\n%s\n"
+                        "The following attributes of the dataframe %s "
+                        "have the wrong dtype:\n%s\n"
                         "They are:\n%s\nbut should be:\n%s",
-                        attr,
                         c.list_name,
                         unmatched.index[unmatched],
-                        c.pnl[attr].dtypes[unmatched],
-                        typ,
+                        c.df.dtypes[dtypes_soll.index[unmatched]],
+                        dtypes_soll[unmatched],
                     )
 
+                # now check varying attributes
+
+                types_soll = c.attrs.loc[c.attrs["varying"], ["typ", "dtype"]]
+
+                for attr, typ, dtype in types_soll.itertuples():
+                    if c.pnl[attr].empty:
+                        continue
+
+                    unmatched = c.pnl[attr].dtypes != dtype
+
+                    if unmatched.any():
+                        logger.warning(
+                            "The following columns of time-varying attribute "
+                            "%s in %s_t have the wrong dtype:\n%s\n"
+                            "They are:\n%s\nbut should be:\n%s",
+                            attr,
+                            c.list_name,
+                            unmatched.index[unmatched],
+                            c.pnl[attr].dtypes[unmatched],
+                            typ,
+                        )
+
         constraint_periods = set(
             self.global_constraints.investment_period.dropna().unique()
         )
