@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-
 """
 Descriptors for component attributes.
 """
@@ -8,7 +7,7 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
@@ -96,8 +95,7 @@ class Dict(dict):
             raise AttributeError(
                 "'Dict' object attribute " "'{0}' is read-only".format(name)
             )
-        else:
-            self[name] = value
+        self[name] = value
 
     def __getattr__(self, item):
         try:
@@ -160,7 +158,6 @@ def get_switchable_as_dense(network, component, attr, snapshots=None, inds=None)
     --------
     >>> get_switchable_as_dense(network, 'Generator', 'p_max_pu')
     """
-
     df = network.df(component)
     pnl = network.pnl(component)
 
@@ -202,7 +199,7 @@ def get_switchable_as_iter(network, component, attr, snapshots, inds=None):
     snapshots : pandas.Index
         Restrict to these snapshots rather than network.snapshots.
     inds : pandas.Index
-        Restrict to these items rather than all of network.{generators,..}.index
+        Restrict to these items rather than all of network.{generators, ..}.index
 
     Returns
     -------
@@ -212,7 +209,6 @@ def get_switchable_as_iter(network, component, attr, snapshots, inds=None):
     --------
     >>> get_switchable_as_iter(network, 'Generator', 'p_max_pu', snapshots)
     """
-
     df = network.df(component)
     pnl = network.pnl(component)
 
@@ -268,9 +264,7 @@ def allocate_series_dataframes(network, series):
     >>> allocate_series_dataframes(network, {'Generator': ['p'],
                                              'Load': ['p']})
     """
-
     for component, attributes in series.items():
-
         df = network.df(component)
         pnl = network.pnl(component)
 
@@ -295,7 +289,7 @@ def free_output_series_dataframes(network, components=None):
 
 def zsum(s, *args, **kwargs):
     """
-    pandas 0.21.0 changes sum() behavior so that the result of applying sum
+    Pandas 0.21.0 changes sum() behavior so that the result of applying sum
     over an empty DataFrame is NaN.
 
     Meant to be set as pd.Series.zsum = zsum.
@@ -369,7 +363,7 @@ def get_active_assets(n, c, investment_period):
         if period not in n.investment_periods:
             raise ValueError("Investment period not in `network.investment_periods`")
         active[period] = n.df(c).eval("build_year <= @period < build_year + lifetime")
-    return pd.DataFrame(active).any(1)
+    return pd.DataFrame(active).any(axis=1)
 
 
 def get_activity_mask(n, c, sns=None, index=None):
