@@ -1,6 +1,4 @@
 # -*- coding: utf-8 -*-
-
-
 """
 Build optimisation problems from PyPSA networks without Pyomo.
 
@@ -11,12 +9,13 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
 import numpy as np
 import pandas as pd
+from deprecation import deprecated
 from numpy import inf
 from packaging.version import Version, parse
 
@@ -29,8 +28,9 @@ from pypsa.descriptors import (
     get_bounds_pu,
     get_extendable_i,
     get_non_extendable_i,
-    nominal_attrs,
 )
+from pypsa.descriptors import get_switchable_as_dense as get_as_dense
+from pypsa.descriptors import nominal_attrs
 from pypsa.linopt import (
     align_with_static_component,
     define_binaries,
@@ -52,7 +52,6 @@ from pypsa.linopt import (
     write_objective,
 )
 from pypsa.pf import _as_snapshots
-from pypsa.pf import get_switchable_as_dense as get_as_dense
 
 agg_group_kwargs = (
     dict(numeric_only=False) if parse(pd.__version__) >= Version("1.3") else {}
@@ -110,15 +109,17 @@ def define_dispatch_for_extendable_and_committable_variables(n, sns, c, attr):
         name of the attribute, e.g. 'p'
     """
     ext_i = get_extendable_i(n, c)
-    if c == "Generator":
-        ext_i = ext_i.union(n.generators.query("committable").index)
+    if c in {"Generator", "Link"}:
+        ext_i = ext_i.union(
+            getattr(n, n.components[c]["list_name"]).query("committable").index
+        )
     if ext_i.empty:
         return
     active = get_activity_mask(n, c, sns)[ext_i] if n._multi_invest else None
     define_variables(n, -inf, inf, c, attr, axes=[sns, ext_i], spec="ext", mask=active)
 
 
-def define_dispatch_for_non_extendable_variables(n, sns, c, attr):
+def define_dispatch_for_non_extendable_variables(n, sns, c, attr, transmission_losses):
     """
     Initializes variables for power dispatch for a given component and a given
     attribute.
@@ -132,8 +133,10 @@ def define_dispatch_for_non_extendable_variables(n, sns, c, attr):
         name of the attribute, e.g. 'p'
     """
     fix_i = get_non_extendable_i(n, c)
-    if c == "Generator":
-        fix_i = fix_i.difference(n.generators.query("committable").index)
+    if c in {"Generator", "Link"}:
+        fix_i = fix_i.difference(
+            getattr(n, n.components[c]["list_name"]).query("committable").index
+        )
     if fix_i.empty:
         return
     nominal_fix = n.df(c)[nominal_attrs[c]][fix_i]
@@ -146,12 +149,17 @@ def define_dispatch_for_non_extendable_variables(n, sns, c, attr):
     kwargs = dict(spec="non_ext", mask=active)
 
     dispatch = define_variables(n, -inf, inf, c, attr, axes=axes, **kwargs)
-    dispatch = linexpr((1, dispatch))
-    define_constraints(n, dispatch, ">=", lower, c, "mu_lower", **kwargs)
-    define_constraints(n, dispatch, "<=", upper, c, "mu_upper", **kwargs)
+    dispatch_lower = linexpr((1, dispatch))
+    dispatch_upper = linexpr((1, dispatch))
+    if c in n.passive_branch_components and not n.df(c).empty and transmission_losses:
+        loss = get_var(n, c, "loss")[fix_i]
+        dispatch_lower = linexpr((1, dispatch), (-1, loss))
+        dispatch_upper = linexpr((1, dispatch), (1, loss))
+    define_constraints(n, dispatch_lower, ">=", lower, c, "mu_lower", **kwargs)
+    define_constraints(n, dispatch_upper, "<=", upper, c, "mu_upper", **kwargs)
 
 
-def define_dispatch_for_extendable_constraints(n, sns, c, attr):
+def define_dispatch_for_extendable_constraints(n, sns, c, attr, transmission_losses):
     """
     Sets power dispatch constraints for extendable devices for a given
     component and a given attribute.
@@ -175,10 +183,16 @@ def define_dispatch_for_extendable_constraints(n, sns, c, attr):
     active = get_activity_mask(n, c, sns)[ext_i] if n._multi_invest else None
     kwargs = dict(spec=attr, mask=active)
 
-    lhs, *axes = linexpr((max_pu, nominal_v), (-1, operational_ext_v), return_axes=True)
+    lhs_upper = [(max_pu, nominal_v), (-1, operational_ext_v)]
+    lhs_lower = [(min_pu, nominal_v), (-1, operational_ext_v)]
+    if c in n.passive_branch_components and not n.df(c).empty and transmission_losses:
+        loss = get_var(n, c, "loss")[ext_i]
+        lhs_upper.append((-1, loss))
+        lhs_lower.append((-1, loss))
+    lhs, *axes = linexpr(*lhs_upper, return_axes=True)
     define_constraints(n, lhs, ">=", rhs, c, "mu_upper", axes=axes, **kwargs)
 
-    lhs, *axes = linexpr((min_pu, nominal_v), (-1, operational_ext_v), return_axes=True)
+    lhs, *axes = linexpr(*lhs_lower, return_axes=True)
     define_constraints(n, lhs, "<=", rhs, c, "mu_lower", axes=axes, **kwargs)
 
 
@@ -198,7 +212,6 @@ def define_fixed_variable_constraints(n, sns, c, attr, pnl=True):
     pnl : bool, default True
         Whether variable which should be fixed is time-dependent
     """
-
     if pnl:
         if attr + "_set" not in n.pnl(c):
             return
@@ -225,26 +238,96 @@ def define_fixed_variable_constraints(n, sns, c, attr, pnl=True):
     set_conref(n, constraints, c, f"mu_{attr}_set")
 
 
+@deprecated(
+    deprecated_in="0.23",
+    removed_in="0.24",
+    details="Use define_unit_commitment_status_variables instead.",
+)
 def define_generator_status_variables(n, sns):
-    c = "Generator"
-    com_i = n.generators.query("committable").index
+    define_unit_commitment_status_variables(n, sns, "Generator")
+
+
+def define_unit_commitment_status_variables(n, sns, c):
+    allowed_c = {"Generator", "Link"}
+    assert c in allowed_c, f"Component {c} must be in {allowed_c}."
+    com_i = n.df(c).query("committable").index
     ext_i = get_extendable_i(n, c)
     if not (ext_i.intersection(com_i)).empty:
         logger.warning(
-            "The following generators have both investment optimisation"
+            f"The following {c} components have both investment optimisation"
             f" and unit commitment:\n\n\t{', '.join((ext_i.intersection(com_i)))}\n\nCurrently PyPSA cannot "
             "do both these functions, so PyPSA is choosing investment optimisation "
-            "for these generators."
+            f"for these {c} components."
         )
         com_i = com_i.difference(ext_i)
     if com_i.empty:
         return
     active = get_activity_mask(n, c, sns)[com_i] if n._multi_invest else None
-    define_binaries(n, (sns, com_i), "Generator", "status", mask=active)
+    define_binaries(n, (sns, com_i), c, "status", mask=active)
 
 
+def define_loss_variables(n, sns, c, transmission_losses):
+    branches_i = n.df(c).index
+    if branches_i.empty or not transmission_losses:
+        return
+    active = get_activity_mask(n, c, sns) if n._multi_invest else None
+    define_variables(n, 0, inf, c, "loss", axes=[sns, branches_i], mask=active)
+
+
+def define_loss_constraints(n, sns, c, transmission_losses):
+    if not transmission_losses or n.df(c).empty:
+        return
+
+    tangents = transmission_losses
+    active = get_activity_mask(n, c, sns) if n._multi_invest else None
+
+    s_max_pu = get_as_dense(n, c, "s_max_pu").loc[sns]
+
+    s_nom_max = n.df(c)["s_nom_max"].where(
+        n.df(c)["s_nom_extendable"], n.df(c)["s_nom"]
+    )
+
+    assert np.isfinite(
+        s_nom_max
+    ).all(), f"Loss approximation requires finite 's_nom_max' for extendable branches:\n {s_nom_max[~np.isfinite(s_nom_max)]}"
+
+    r_pu_eff = n.df(c)["r_pu_eff"]
+
+    rhs = r_pu_eff * (s_max_pu * s_nom_max) ** 2
+
+    loss = get_var(n, c, "loss")
+    lhs = linexpr((1, loss))
+
+    define_constraints(n, lhs, "<=", rhs, c, "loss_upper", mask=active)
+
+    flow = get_var(n, c, "s")
+
+    for k in range(1, tangents + 1):
+        p_k = k / tangents * s_max_pu * s_nom_max
+        loss_k = r_pu_eff * p_k**2
+        slope_k = 2 * r_pu_eff * p_k
+        offset_k = loss_k - slope_k * p_k
+
+        for sign in [-1, 1]:
+            lhs = linexpr((1, loss), (sign * slope_k, flow))
+
+            define_constraints(
+                n, lhs, ">=", offset_k, c, f"loss-tangents-{k}-{sign}", mask=active
+            )
+
+
+@deprecated(
+    deprecated_in="0.23",
+    removed_in="0.24",
+    details="Use define_unit_commitment_constraints instead.",
+)
 def define_committable_generator_constraints(n, sns):
-    c, attr = "Generator", "status"
+    define_unit_commitment_constraints(n, sns, "Generator")
+
+
+def define_unit_commitment_constraints(n, sns, c):
+    allowed_c = {"Generator", "Link"}
+    assert c in allowed_c, f"Component {c} must be in {allowed_c}."
     com_i = n.df(c).query("committable and not p_nom_extendable").index
     if com_i.empty:
         return
@@ -253,15 +336,16 @@ def define_committable_generator_constraints(n, sns):
     lower = min_pu.mul(nominal)
     upper = max_pu.mul(nominal)
 
-    status = get_var(n, c, attr)
+    status = get_var(n, c, "status")
+
     p = get_var(n, c, "p")[com_i]
 
     lhs = linexpr((lower, status), (-1, p))
     active = get_activity_mask(n, c, sns)[com_i] if n._multi_invest else None
-    define_constraints(n, lhs, "<=", 0, "Generators", "committable_lb", mask=active)
+    define_constraints(n, lhs, "<=", 0, c, "committable_lb", mask=active)
 
     lhs = linexpr((upper, status), (-1, p))
-    define_constraints(n, lhs, ">=", 0, "Generators", "committable_ub", mask=active)
+    define_constraints(n, lhs, ">=", 0, c, "committable_ub", mask=active)
 
 
 def define_ramp_limit_constraints(n, sns, c):
@@ -385,7 +469,6 @@ def define_ramp_limit_constraints(n, sns, c):
 def define_nominal_constraints_per_bus_carrier(n, sns):
     for carrier in n.carriers.index:
         for bound, sense in [("max", "<="), ("min", ">=")]:
-
             col = f"nom_{bound}_{carrier}"
             if col not in n.buses.columns:
                 continue
@@ -417,11 +500,12 @@ def define_nominal_constraints_per_bus_carrier(n, sns):
             define_constraints(n, lhs, sense, rhs, "Bus", "mu_" + col)
 
 
-def define_nodal_balance_constraints(n, sns):
+def define_nodal_balance_constraints(n, sns, transmission_losses):
     """
     Defines nodal balance constraint.
     """
 
+    #
     def bus_injection(c, attr, groupcol="bus", sign=1):
         # additional sign only necessary for branches in reverse direction
         if "sign" in n.df(c):
@@ -447,6 +531,14 @@ def define_nodal_balance_constraints(n, sns):
     ]
     args = [arg for arg in args if not n.df(arg[0]).empty]
 
+    if transmission_losses:
+        args.extend(
+            [
+                ["Line", "loss", "bus0", -0.5],
+                ["Line", "loss", "bus1", -0.5],
+            ]
+        )
+
     if not n.links.empty:
         for i in additional_linkports(n):
             eff = get_as_dense(n, "Link", f"efficiency{i}", sns)
@@ -508,7 +600,7 @@ def define_kirchhoff_constraints(n, sns):
             weightings = branches.x_pu_eff if carrier == "AC" else branches.r_pu_eff
             C_weighted = 1e5 * C.mul(weightings, axis=0)
             cycle_sum = C_weighted.apply(cycle_flow, sns=period)
-            snapshots = sns if period == None else sns[sns.get_loc(period)]
+            snapshots = sns if period is None else sns[sns.get_loc(period)]
             cycle_sum.set_index(snapshots, inplace=True)
 
             con = write_constraint(n, cycle_sum, "=", 0)
@@ -544,9 +636,9 @@ def define_storage_unit_constraints(n, sns):
     # elapsed hours
     eh = expand_series(n.snapshot_weightings.stores[sns], sus_i)
     # efficiencies
-    eff_stand = expand_series(1 - n.df(c).standing_loss, sns).T.pow(eh)
-    eff_dispatch = expand_series(n.df(c).efficiency_dispatch, sns).T
-    eff_store = expand_series(n.df(c).efficiency_store, sns).T
+    eff_stand = (1 - get_as_dense(n, c, "standing_loss", sns)).pow(eh)
+    eff_dispatch = get_as_dense(n, c, "efficiency_dispatch", sns)
+    eff_store = get_as_dense(n, c, "efficiency_store", sns)
 
     soc = get_var(n, c, "state_of_charge")
 
@@ -656,7 +748,7 @@ def define_store_constraints(n, sns):
 
     # elapsed hours
     eh = expand_series(n.snapshot_weightings.stores[sns], stores_i)  # elapsed hours
-    eff_stand = expand_series(1 - n.df(c).standing_loss, sns).T.pow(eh)
+    eff_stand = (1 - get_as_dense(n, c, "standing_loss", sns)).pow(eh)
 
     e = get_var(n, c, "e")
 
@@ -739,6 +831,13 @@ def define_growth_limit(n, sns, c, attr):
     ext_i = get_extendable_i(n, c)
     if "carrier" not in n.df(c) or n.df(c).empty:
         return
+
+    if (n.carriers.max_relative_growth > 0).any():
+        logger.warning(
+            "Max relative growth is not implemented for the native pypsa optimization framework. "
+            "Use the linopy framework with `n.optimize` instead."
+        )
+
     with_limit = n.carriers.query("max_growth != inf").index
     limit_i = n.df(c).query("carrier in @with_limit").index.intersection(ext_i)
     if limit_i.empty:
@@ -781,7 +880,6 @@ def define_global_constraints(n, sns):
         technology in a certain region. Currently, only the
         capacities of extendable generators have to be below the set limit.
     """
-
     if n._multi_invest:
         period_weighting = n.investment_period_weightings["years"]
         weightings = n.snapshot_weightings.mul(period_weighting, level=0, axis=0).loc[
@@ -816,11 +914,9 @@ def define_global_constraints(n, sns):
         # generators
         gens = n.generators.query("carrier in @emissions.index")
         if not gens.empty:
-            em_pu = gens.carrier.map(emissions) / gens.efficiency
-            em_pu = (
-                weightings["generators"].to_frame("weightings")
-                @ em_pu.to_frame("weightings").T
-            ).loc[period]
+            efficiency = get_as_dense(n, "Generator", "efficiency", inds=gens.index)
+            em_pu = gens.carrier.map(emissions) / efficiency
+            em_pu = em_pu.multiply(weightings.generators, axis=0).loc[period]
             p = get_var(n, "Generator", "p").loc[sns, gens.index].loc[period]
 
             vals = linexpr((em_pu, p), as_pandas=False)
@@ -867,7 +963,10 @@ def define_global_constraints(n, sns):
     glcs = n.global_constraints.query(
         "type == " '"transmission_volume_expansion_limit"'
     )
-    substr = lambda s: re.sub(r"[\[\]\(\)]", "", s)
+
+    def substr(s):
+        return re.sub("[\\[\\]\\(\\)]", "", s)
+
     for name, glc in glcs.iterrows():
         car = [substr(c.strip()) for c in glc.carrier_attribute.split(",")]
         lhs = ""
@@ -934,7 +1033,9 @@ def define_global_constraints(n, sns):
 
     # (4) tech_capacity_expansion_limit
     # TODO: Generalize to carrier capacity expansion limit (i.e. also for stores etc.)
-    substr = lambda s: re.sub(r"[\[\]\(\)]", "", s)
+    def substr(s):
+        return re.sub("[\\[\\]\\(\\)]", "", s)
+
     glcs = n.global_constraints.query("type == " '"tech_capacity_expansion_limit"')
     c, attr = "Generator", "p_nom"
 
@@ -972,7 +1073,6 @@ def define_objective(n, sns):
     """
     Defines and writes out the objective function.
     """
-
     if n._multi_invest:
         period_weighting = n.investment_period_weightings.objective[
             sns.unique("period")
@@ -1013,7 +1113,7 @@ def define_objective(n, sns):
     for c, attr in lookup.query("marginal_cost").index:
         cost = (
             get_as_dense(n, c, "marginal_cost", sns)
-            .loc[:, lambda ds: (ds != 0).all()]
+            .loc[:, lambda ds: (ds != 0).any()]
             .mul(weighting, axis=0)
         )
         if cost.empty:
@@ -1050,6 +1150,7 @@ def prepare_lopf(
     skip_objective=False,
     extra_functionality=None,
     solver_dir=None,
+    transmission_losses=0,
 ):
     """
     Sets up the linear problem and writes it out to a lp file.
@@ -1082,36 +1183,45 @@ def prepare_lopf(
     n.bounds_f = open(bounds_fn, mode="w")
     n.binaries_f = open(binaries_fn, mode="w")
 
-    n.objective_f.write("\* LOPF *\n\nmin\nobj:\n")
+    n.objective_f.write("\ LOPF \n\nmin\nobj:\n")
     n.constraints_f.write("\n\ns.t.\n\n")
     n.bounds_f.write("\nbounds\n")
     n.binaries_f.write("\nbinary\n")
 
+    for c in n.passive_branch_components:
+        define_loss_variables(n, snapshots, c, transmission_losses)
     for c, attr in lookup.query("nominal and not handle_separately").index:
         define_nominal_for_extendable_variables(n, c, attr)
         # define constraint for newly installed capacity per investment period
         define_growth_limit(n, snapshots, c, attr)
         # define_fixed_variable_constraints(n, snapshots, c, attr, pnl=False)
     for c, attr in lookup.query("not nominal and not handle_separately").index:
-        define_dispatch_for_non_extendable_variables(n, snapshots, c, attr)
+        define_dispatch_for_non_extendable_variables(
+            n, snapshots, c, attr, transmission_losses
+        )
         define_dispatch_for_extendable_and_committable_variables(n, snapshots, c, attr)
         align_with_static_component(n, c, attr)
-        define_dispatch_for_extendable_constraints(n, snapshots, c, attr)
+        define_dispatch_for_extendable_constraints(
+            n, snapshots, c, attr, transmission_losses
+        )
         # define_fixed_variable_constraints(n, snapshots, c, attr)
-    define_generator_status_variables(n, snapshots)
+    define_unit_commitment_status_variables(n, snapshots, c="Generator")
+    define_unit_commitment_status_variables(n, snapshots, c="Link")
     define_nominal_constraints_per_bus_carrier(n, snapshots)
 
     # consider only state_of_charge_set for the moment
     define_fixed_variable_constraints(n, snapshots, "StorageUnit", "state_of_charge")
     define_fixed_variable_constraints(n, snapshots, "Store", "e")
 
-    define_committable_generator_constraints(n, snapshots)
-    define_ramp_limit_constraints(n, snapshots, c="Generator")
-    define_ramp_limit_constraints(n, snapshots, c="Link")
+    for c in {"Generator", "Link"}:
+        define_unit_commitment_constraints(n, snapshots, c)
+        define_ramp_limit_constraints(n, snapshots, c)
     define_storage_unit_constraints(n, snapshots)
     define_store_constraints(n, snapshots)
     define_kirchhoff_constraints(n, snapshots)
-    define_nodal_balance_constraints(n, snapshots)
+    define_nodal_balance_constraints(n, snapshots, transmission_losses)
+    for c in n.passive_branch_components:
+        define_loss_constraints(n, snapshots, c, transmission_losses)
     define_global_constraints(n, snapshots)
     if skip_objective:
         logger.info(
@@ -1190,6 +1300,8 @@ def assign_solution(
             if c in n.passive_branch_components and attr == "s":
                 set_from_frame(pnl, "p0", values)
                 set_from_frame(pnl, "p1", -values)
+            if c in n.passive_branch_components and attr == "loss":
+                set_from_frame(pnl, "loss", values)
             elif c == "Link" and attr == "p":
                 set_from_frame(pnl, "p0", values)
                 for i in ["1"] + additional_linkports(n):
@@ -1226,7 +1338,7 @@ def assign_solution(
         n.pnl(c)["p"] = n.pnl(c)["p_dispatch"] - n.pnl(c)["p_store"]
 
     # duals
-    if keep_shadowprices == False:
+    if keep_shadowprices is False:
         keep_shadowprices = []
 
     sp = n.constraints.index
@@ -1314,7 +1426,9 @@ def assign_solution(
     for i in additional_linkports(n):
         ca.append(("Link", f"p{i}", f"bus{i}"))
 
-    sign = lambda c: n.df(c).sign if "sign" in n.df(c) else -1  # sign for 'Link'
+    def sign(c):
+        return n.df(c).sign if "sign" in n.df(c) else -1  # sign for 'Link'
+
     n.buses_t.p = (
         pd.concat(
             [
@@ -1353,6 +1467,7 @@ def network_lopf(
     skip_pre=False,
     extra_postprocessing=None,
     formulation="kirchhoff",
+    transmission_losses=0,
     keep_references=False,
     keep_files=False,
     keep_shadowprices=["Bus", "Line", "Transformer", "Link", "GlobalConstraint"],
@@ -1385,10 +1500,15 @@ def network_lopf(
         construction, e.g. .lp file - useful for debugging
     formulation : string
         Formulation of the linear power flow equations to use; must be
-        one of ["angles","cycles","kirchhoff","ptdf"]
+        one of ["angles", "cycles", "kirchhoff", "ptdf"]
+    transmission_losses : int, default 0
+        Whether an approximation of transmission losses should be included
+        in the linearised power flow formulation. A passed number will denote
+        the number of tangents used for the piecewise linear approximation.
+        Defaults to 0 which ignores losses.
     extra_functionality : callable function
         This function must take two arguments
-        `extra_functionality(network,snapshots)` and is called after
+        `extra_functionality(network, snapshots)` and is called after
         the model building is complete, but before it is sent to the
         solver. It allows the user to
         add/change constraints and add/change the objective function.
@@ -1399,7 +1519,7 @@ def network_lopf(
         objective has to be defined via extra_functionality.
     extra_postprocessing : callable function
         This function must take three arguments
-        `extra_postprocessing(network,snapshots,duals)` and is called after
+        `extra_postprocessing(network, snapshots, duals)` and is called after
         the model has solved and the results are extracted. It allows the user
         to extract further information about the solution, such as additional
         shadow prices.
@@ -1431,7 +1551,7 @@ def network_lopf(
     if formulation != "kirchhoff":
         raise NotImplementedError("Only the kirchhoff formulation is supported")
 
-    if n.generators.committable.any():
+    if n.generators.committable.any() or n.links.committable.any():
         logger.warning(
             "Unit commitment is not yet completely implemented for "
             "optimising without pyomo. Thus minimum up time, minimum down time, "
@@ -1454,11 +1574,17 @@ def network_lopf(
 
     logger.info("Prepare linear problem")
     fdp, problem_fn = prepare_lopf(
-        n, snapshots, keep_files, skip_objective, extra_functionality, solver_dir
+        n,
+        snapshots,
+        keep_files,
+        skip_objective,
+        extra_functionality,
+        solver_dir,
+        transmission_losses,
     )
     fds, solution_fn = mkstemp(prefix="pypsa-solve", suffix=".sol", dir=solver_dir)
 
-    if warmstart == True:
+    if warmstart is True:
         warmstart = n.basis_fn
         logger.info("Solve linear problem using warmstart")
     else:
