@@ -1,7 +1,5 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-
-
 """
 Tools for fast Linear Problem file writing. This module contains.
 
@@ -19,7 +17,7 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
@@ -51,7 +49,7 @@ def define_variables(n, lower, upper, name, attr="", axes=None, spec="", mask=No
     bound(s). The variables are stored in the network object under n.vars with
     key of the variable name. If multiple variables are defined at ones, at
     least one of lower and upper has to be an array (including pandas) of
-    `shape > (1,)` or axes have to define the dimensions of the variables.
+    `shape > (1, )` or axes have to define the dimensions of the variables.
 
     Parameters
     ----------
@@ -432,10 +430,17 @@ def to_pandas(array, *axes):
     return pd.Series(array, *axes) if array.ndim == 1 else pd.DataFrame(array, *axes)
 
 
-_to_float_str = lambda f: "%+f" % f
+def _to_float_str(f):
+    return "%+f" % f
+
+
 _v_to_float_str = np.vectorize(_to_float_str, otypes=[object])
 
-_to_int_str = lambda d: "%d" % d
+
+def _to_int_str(d):
+    return "%d" % d
+
+
 _v_to_int_str = np.vectorize(_to_int_str, otypes=[object])
 
 
@@ -452,8 +457,7 @@ def _str_array(array, integer_string=False):
             array = np.nan_to_num(array, False, -1)
             return _v_to_int_str(array)
         return _v_to_float_str(array)
-    else:
-        return array
+    return array
 
 
 def join_exprs(df):
@@ -589,8 +593,7 @@ def get_sol(n, name, attr=""):
     pnl = n.solutions.at[(name, attr), "pnl"]
     if n.solutions.at[(name, attr), "in_comp"]:
         return n.pnl(name)[attr] if pnl else n.df(name)[attr + "_opt"]
-    else:
-        return n.sols[name].pnl[attr] if pnl else n.sols[name].df[attr]
+    return n.sols[name].pnl[attr] if pnl else n.sols[name].df[attr]
 
 
 def get_dual(n, name, attr=""):
@@ -615,8 +618,7 @@ def get_dual(n, name, attr=""):
     pnl = n.dualvalues.at[(name, attr), "pnl"]
     if n.dualvalues.at[(name, attr), "in_comp"]:
         return n.pnl(name)[attr] if pnl else n.df(name)[attr]
-    else:
-        return n.duals[name].pnl[attr] if pnl else n.duals[name].df[attr]
+    return n.duals[name].pnl[attr] if pnl else n.duals[name].df[attr]
 
 
 # =============================================================================
@@ -666,7 +668,7 @@ def run_and_read_highs(
     Now when typing ``highs`` in the terminal you should see something like ::
         Running HiGHS 1.1.1 [date: 2021-11-14, git hash: 95342daa]
 
-    The function reads and execute (i.e. subprocess.Popen,...) terminal
+    The function reads and execute (i.e. subprocess.Popen, ...) terminal
     commands of the solver. Meaning the command can be also executed at your
     command window/terminal if HiGHs is installed. Executing the commands on
     your local terminal helps to identify the raw outputs that are useful for
@@ -720,7 +722,7 @@ def run_and_read_highs(
 
     """
     logger.warning(
-        "The HiGHS solver can potentially solve towards variables that slightly deviate from Gurobi,cbc,glpk"
+        "The HiGHS solver can potentially solve towards variables that slightly deviate from Gurobi, cbc, glpk"
     )
     options_fn = "highs_options.txt"
     default_dict = {
@@ -762,7 +764,7 @@ def run_and_read_highs(
         # Function that reads line by line the command window
         while True:
             out = process.stdout.readline(1)
-            if out == "" and process.poll() != None:
+            if out == "" and process.poll() is not None:
                 break
             if out != "":
                 yield out
@@ -793,7 +795,10 @@ def run_and_read_highs(
     trimed_sol_fn = re.sub(rb"\*\*\s+", b"", f.read())
     f.close()
 
-    sol = pd.read_csv(io.BytesIO(trimed_sol_fn), header=[1], sep=r"\s+")
+    sol = pd.read_fwf(io.BytesIO(trimed_sol_fn), header=[1])
+
+    sol = sol.iloc[:-2, :]  # last two rows are model info: status and objective value
+
     row_no = sol[sol["Index"] == "Rows"].index[0]
     sol = sol.drop(row_no + 1)  # Removes header line after "Rows"
     sol_rows = sol[(sol.index > row_no)]
@@ -825,8 +830,8 @@ def run_and_read_cbc(
     """
     with open(problem_fn, "rb") as f:
         for str in f.readlines():
-            assert ("> " in str.decode("utf-8")) == False, ">, must be" "changed to >="
-            assert ("< " in str.decode("utf-8")) == False, "<, must be" "changed to <="
+            assert ("> " in str.decode("utf-8")) is False, ">, must be" "changed to >="
+            assert ("< " in str.decode("utf-8")) is False, "<, must be" "changed to <="
 
     # printingOptions is about what goes in solution file
     command = f"cbc -printingOptions all -import {problem_fn} "
@@ -1055,8 +1060,8 @@ def run_and_read_gurobi(
     solver. If the solution is successful it returns variable solutions and
     constraint dual values. Gurobipy must be installed for using this function.
 
-    For more information on solver options:
-    https://www.gurobi.com/documentation/{gurobi_verion}/refman/parameter_descriptions.html
+    For more information on solver options, see
+    <https://www.gurobi.com/documentation/{gurobi_verion}/refman/parameter_descriptions.html>_
     """
     if find_spec("gurobipy") is None:
         raise ModuleNotFoundError(
@@ -1140,8 +1145,8 @@ def run_and_read_xpress(
     constraint dual values. The xpress module must be installed for using this
     function.
 
-    For more information on solver options:
-    https://www.fico.com/fico-xpress-optimization/docs/latest/solver/GUID-ACD7E60C-7852-36B7-A78A-CED0EA291CDD.html
+    For more information on solver options, see
+    <https://www.fico.com/fico-xpress-optimization/docs/latest/solver/GUID-ACD7E60C-7852-36B7-A78A-CED0EA291CDD.html>_
     """
 
     import xpress
@@ -1155,12 +1160,17 @@ def run_and_read_xpress(
         m.setlogfile(solver_logfile)
 
     if warmstart:
-        m.readbasis(warmstart)
+        try:
+            m.readbasis(n.basis_fn)
+            logger.info("Model basis loaded")
+        except:
+            logger.info("No model basis loaded")
+            pass
 
     m.solve()
 
     if store_basis:
-        n.basis_fn = solution_fn.replace(".sol", ".bas")
+        n.basis_fn = solution_fn.replace(".sol", ".bss")
         try:
             m.writebasis(n.basis_fn)
         except:
