@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-
 """
 Functions for computing network clusters.
 """
@@ -8,7 +7,7 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
@@ -34,8 +33,7 @@ def _normed(s):
     tot = s.sum()
     if tot == 0:
         return 1.0
-    else:
-        return s / tot
+    return s / tot
 
 
 def _flatten_multiindex(m, join=" "):
@@ -122,6 +120,9 @@ def aggregategenerators(
         for attr, df in network.generators_t.items():
             pnl_gens_agg_b = df.columns.to_series().map(gens_agg_b)
             df_agg = df.loc[:, pnl_gens_agg_b]
+            # If there are any generators to aggregate, do so and put
+            # the time-varying data for all generators (aggregated and
+            # non-aggregated) into `new_pnl`.
             if not df_agg.empty:
                 if attr == "p_max_pu":
                     df_agg = df_agg.multiply(weighting.loc[df_agg.columns], axis=1)
@@ -130,6 +131,11 @@ def aggregategenerators(
                 new_pnl[attr] = pd.concat(
                     [df.loc[:, ~pnl_gens_agg_b], pnl_df], axis=1, sort=False
                 )
+            # Even if no generators are aggregated, we still need to
+            # put the time-varying data for all non-aggregated
+            # generators into `new_pnl`.
+            elif not df.empty:
+                new_pnl[attr] = df
 
     return new_df, new_pnl
 
@@ -137,7 +143,6 @@ def aggregategenerators(
 def aggregateoneport(
     network, busmap, component, with_time=True, custom_strategies=dict()
 ):
-
     if network.df(component).empty:
         return network.df(component), network.pnl(component)
 
@@ -266,7 +271,6 @@ def aggregatelines(network, buses, interlines, line_length_factor=1.0, with_time
     }
 
     def aggregatelinegroup(l):
-
         # l.name is a tuple of the groupby index (bus0_s, bus1_s)
         length_s = (
             haversine_pts(
@@ -301,6 +305,8 @@ def aggregatelines(network, buses, interlines, line_length_factor=1.0, with_time
             sub_network=consense["sub_network"](l["sub_network"]),
             v_ang_min=l["v_ang_min"].max(),
             v_ang_max=l["v_ang_max"].min(),
+            lifetime=(l["lifetime"] * _normed(l["s_nom"])).sum(),
+            build_year=(l["build_year"] * _normed(l["s_nom"])).sum(),
         )
         data.update((f, consense[f](l[f])) for f in columns.difference(data))
         return pd.Series(data, index=[f for f in l.columns if f in columns])
@@ -344,14 +350,19 @@ def get_buses_linemap_and_lines(
     lines, linemap_p, linemap_n, linemap, lines_t = aggregatelines(
         network, buses, interlines, line_length_factor, with_time
     )
+    # network can be reduced to a set of isolated nodes in course of clustering (e.g. Rwanda)
+    if lines.empty:
+        lines_res = lines.drop(columns=["bus0_s", "bus1_s"]).reset_index(drop=True)
+    else:
+        lines_res = lines.reset_index().rename(
+            columns={"bus0_s": "bus0", "bus1_s": "bus1"}, copy=False
+        )
     return (
         buses,
         linemap,
         linemap_p,
         linemap_n,
-        lines.reset_index()
-        .rename(columns={"bus0_s": "bus0", "bus1_s": "bus1"}, copy=False)
-        .set_index("name"),
+        lines_res.set_index("name"),
         lines_t,
     )
 
@@ -375,7 +386,6 @@ def get_clustering_from_busmap(
     one_port_strategies=dict(),
     generator_strategies=dict(),
 ):
-
     buses, linemap, linemap_p, linemap_n, lines, lines_t = get_buses_linemap_and_lines(
         network, busmap, line_length_factor, bus_strategies, with_time
     )
@@ -486,7 +496,7 @@ def busmap_by_kmeans(network, bus_weightings, n_clusters, buses_i=None, **kwargs
     Parameters
     ----------
     network : pypsa.Network
-        The buses must have coordinates x,y.
+        The buses must have coordinates x, y.
     bus_weightings : pandas.Series
         Series of integer weights for buses, indexed by bus names.
     n_clusters : int
@@ -502,7 +512,6 @@ def busmap_by_kmeans(network, bus_weightings, n_clusters, buses_i=None, **kwargs
         Mapping of network.buses to k-means clusters (indexed by
         non-negative integers).
     """
-
     if find_spec("sklearn") is None:
         raise ModuleNotFoundError(
             "Optional dependency 'sklearn' not found."
@@ -522,6 +531,7 @@ def busmap_by_kmeans(network, bus_weightings, n_clusters, buses_i=None, **kwargs
         bus_weightings.reindex(buses_i).astype(int), axis=0
     )
 
+    kwargs.setdefault("n_init", "auto")
     kmeans = KMeans(init="k-means++", n_clusters=n_clusters, **kwargs)
 
     kmeans.fit(points)
@@ -547,7 +557,7 @@ def kmeans_clustering(
     Parameters
     ----------
     network : pypsa.Network
-        The buses must have coordinates x,y.
+        The buses must have coordinates x, y.
     bus_weightings : pandas.Series
         Series of integer weights for buses, indexed by bus names.
     n_clusters : int
@@ -652,10 +662,11 @@ def busmap_by_hac(
         :, buses_x
     ]
 
+    # TODO: maybe change the deprecated argument 'affinity' to 'metric'
     labels = HAC(
         n_clusters=n_clusters,
         connectivity=A,
-        affinity=affinity,
+        metric=affinity,
         linkage=linkage,
         **kwargs,
     ).fit_predict(feature)
@@ -682,7 +693,7 @@ def hac_clustering(
     Parameters
     ----------
     network : pypsa.Network
-        The buses must have coordinates x,y.
+        The buses must have coordinates x, y.
     buses_i: None | pandas.Index, default=None
         Subset of buses to cluster. If None, all buses are considered.
     branch_components: List, default=["Line", "Link"]
@@ -717,7 +728,14 @@ def hac_clustering(
     """
 
     busmap = busmap_by_hac(
-        network, n_clusters, buses_i, branch_components, feature, **kwargs
+        network,
+        n_clusters,
+        buses_i,
+        branch_components,
+        feature,
+        affinity,
+        linkage,
+        **kwargs,
     )
 
     return get_clustering_from_busmap(
@@ -752,7 +770,6 @@ def busmap_by_greedy_modularity(network, n_clusters, buses_i=None):
        "Finding community structure in very large networks."
        Physical Review E 70(6), 2004.
     """
-
     if parse(nx.__version__) < Version("2.8"):
         raise NotImplementedError(
             "The fuction `busmap_by_greedy_modularity` requires `networkx>=2.8`, "
@@ -844,7 +861,6 @@ def busmap_by_stubs(network, matching_attrs=None):
         Mapping of network.buses to k-means clusters (indexed by
         non-negative integers).
     """
-
     busmap = pd.Series(network.buses.index, network.buses.index)
 
     G = network.graph()
