@@ -6,10 +6,10 @@ Define optimisation constraints from PyPSA networks with Linopy.
 import logging
 
 import pandas as pd
-from linopy.expressions import LinearExpression, merge
-from numpy import arange, cumsum, inf, nan, roll
+from linopy import LinearExpression, Variable, merge
+from numpy import arange, cumsum, inf, isfinite, nan, roll
 from scipy import sparse
-from xarray import DataArray, Dataset, zeros_like
+from xarray import DataArray, Dataset, concat
 
 from pypsa.descriptors import (
     additional_linkports,
@@ -20,11 +20,14 @@ from pypsa.descriptors import (
 from pypsa.descriptors import get_switchable_as_dense as get_as_dense
 from pypsa.descriptors import nominal_attrs
 from pypsa.optimization.common import reindex
+from pypsa.optimization.compat import define_constraints, get_var, linexpr
 
 logger = logging.getLogger(__name__)
 
 
-def define_operational_constraints_for_non_extendables(n, sns, c, attr):
+def define_operational_constraints_for_non_extendables(
+    n, sns, c, attr, transmission_losses
+):
     """
     Sets power dispatch constraints for non-extendable and non-commitable
     assets for a given component and a given attribute.
@@ -52,12 +55,23 @@ def define_operational_constraints_for_non_extendables(n, sns, c, attr):
 
     active = get_activity_mask(n, c, sns, fix_i) if n._multi_invest else None
 
-    dispatch = reindex(n.model[f"{c}-{attr}"], c, fix_i)
-    n.model.add_constraints(dispatch, ">=", lower, f"{c}-fix-{attr}-lower", active)
-    n.model.add_constraints(dispatch, "<=", upper, f"{c}-fix-{attr}-upper", active)
+    dispatch_lower = reindex(n.model[f"{c}-{attr}"], c, fix_i)
+    dispatch_upper = reindex(n.model[f"{c}-{attr}"], c, fix_i)
+    if c in n.passive_branch_components and transmission_losses:
+        loss = reindex(n.model[f"{c}-loss"], c, fix_i)
+        dispatch_lower = (1, dispatch_lower), (-1, loss)
+        dispatch_upper = (1, dispatch_upper), (1, loss)
+    n.model.add_constraints(
+        dispatch_lower, ">=", lower, f"{c}-fix-{attr}-lower", active
+    )
+    n.model.add_constraints(
+        dispatch_upper, "<=", upper, f"{c}-fix-{attr}-upper", active
+    )
 
 
-def define_operational_constraints_for_extendables(n, sns, c, attr):
+def define_operational_constraints_for_extendables(
+    n, sns, c, attr, transmission_losses
+):
     """
     Sets power dispatch constraints for extendable devices for a given
     component and a given attribute.
@@ -83,17 +97,23 @@ def define_operational_constraints_for_extendables(n, sns, c, attr):
 
     active = get_activity_mask(n, c, sns, ext_i) if n._multi_invest else None
 
-    lhs = (1, dispatch), (-min_pu, capacity)
-    n.model.add_constraints(lhs, ">=", 0, f"{c}-ext-{attr}-lower", active)
+    lhs_lower = (1, dispatch), (-min_pu, capacity)
+    lhs_upper = (1, dispatch), (-max_pu, capacity)
+    if c in n.passive_branch_components and transmission_losses:
+        loss = reindex(n.model[f"{c}-loss"], c, ext_i)
+        lhs_upper += ((1, loss),)
+        lhs_lower += ((1, loss),)
 
-    lhs = (1, dispatch), (-max_pu, capacity)
-    n.model.add_constraints(lhs, "<=", 0, f"{c}-ext-{attr}-upper", active)
+    n.model.add_constraints(lhs_lower, ">=", 0, f"{c}-ext-{attr}-lower", active)
+    n.model.add_constraints(lhs_upper, "<=", 0, f"{c}-ext-{attr}-upper", active)
 
 
 def define_operational_constraints_for_committables(n, sns, c):
     """
     Sets power dispatch constraints for committable devices for a given
-    component and a given attribute.
+    component and a given attribute. The linearized approximation of the unit
+    commitment problem is inspired by Hua et al. (2017) DOI:
+    10.1109/TPWRS.2017.2735026.
 
     Parameters
     ----------
@@ -108,133 +128,147 @@ def define_operational_constraints_for_committables(n, sns, c):
     if com_i.empty:
         return
 
-    nominal = DataArray(n.df(c)[nominal_attrs[c]].reindex(com_i))
-    min_pu, max_pu = map(DataArray, get_bounds_pu(n, c, sns, com_i, "p"))
-    lower = min_pu * nominal
-    upper = max_pu * nominal
-
+    # variables
     status = n.model[f"{c}-status"]
+    start_up = n.model[f"{c}-start_up"]
+    shut_down = n.model[f"{c}-shut_down"]
+    status_diff = status - status.shift(snapshot=1)
     p = reindex(n.model[f"{c}-p"], c, com_i)
     active = get_activity_mask(n, c, sns, com_i) if n._multi_invest else None
 
-    lhs = (1, p), (-lower, status)
-    n.model.add_constraints(lhs, ">=", 0, f"{c}-com-p-lower", active)
+    # parameters
+    nominal = DataArray(n.df(c)[nominal_attrs[c]].reindex(com_i))
+    min_pu, max_pu = map(DataArray, get_bounds_pu(n, c, sns, com_i, "p"))
+    lower_p = min_pu * nominal
+    upper_p = max_pu * nominal
+    min_up_time_set = n.df(c).min_up_time[com_i]
+    min_down_time_set = n.df(c).min_down_time[com_i]
+    ramp_up_limit = nominal * n.df(c).ramp_limit_up[com_i].fillna(1)
+    ramp_down_limit = nominal * n.df(c).ramp_limit_down[com_i].fillna(1)
+    ramp_start_up = nominal * n.df(c).ramp_limit_start_up[com_i]
+    ramp_shut_down = nominal * n.df(c).ramp_limit_shut_down[com_i]
+    up_time_before_set = n.df(c)["up_time_before"].reindex(com_i)
+    down_time_before_set = n.df(c)["down_time_before"].reindex(com_i)
+    initially_up = up_time_before_set.astype(bool)
+    initially_down = down_time_before_set.astype(bool)
 
-    lhs = (1, p), (-upper, status)
-    n.model.add_constraints(lhs, "<=", 0, f"{c}-com-p-upper", active)
-
-    start_i = n.snapshots.get_loc(sns[0])
-    status_diff = status - status.shift(snapshot=1)
-    min_up_time = n.df(c).min_up_time[com_i]
-    min_down_time = n.df(c).min_down_time[com_i]
-    start_up_cost = n.df(c).start_up_cost
-    shut_down_cost = n.df(c).shut_down_cost
-
-    if min_up_time.sum() + start_up_cost.sum():
-        # find out how long the generator has been up before snapshots
-        up_time_before_set = n.df(c)["up_time_before"].reindex(com_i)
-        until_start = n.pnl(c).status.iloc[:start_i][::-1].reindex(columns=com_i)
-        ref = range(1, len(until_start) + 1)
-        up_time_before = until_start[until_start.cumsum().eq(ref, axis=0)].sum()
-        up_time_before = up_time_before.clip(
-            upper=min_up_time, lower=up_time_before_set
+    # check if there are status calculated/fixed before given sns interval
+    if sns[0] != n.snapshots[0]:
+        start_i = n.snapshots.get_loc(sns[0])
+        # get generators which are online until the first regarded snapshot
+        until_start_up = n.pnl(c).status.iloc[:start_i][::-1].reindex(columns=com_i)
+        ref = range(1, len(until_start_up) + 1)
+        up_time_before = until_start_up[until_start_up.cumsum().eq(ref, axis=0)].sum()
+        up_time_before_set = up_time_before.clip(
+            upper=min_up_time_set, lower=up_time_before_set
+        )
+        # get number of snapshots for generators which are offline before the first regarded snapshot
+        until_start_down = ~until_start_up.astype(bool)
+        ref = range(1, len(until_start_down) + 1)
+        down_time_before = until_start_down[
+            until_start_down.cumsum().eq(ref, axis=0)
+        ].sum()
+        down_time_before_set = down_time_before.clip(
+            upper=min_down_time_set, lower=down_time_before_set
         )
 
-        initially_up = up_time_before.astype(bool)
-        must_stay_up = (min_up_time - up_time_before).clip(lower=0, upper=len(sns))
+    # lower dispatch level limit
+    lhs = (1, p), (-lower_p, status)
+    n.model.add_constraints(lhs, ">=", 0, f"{c}-com-p-lower", mask=active)
 
-    if min_up_time.sum():
-        if initially_up.any():
-            ref = pd.DataFrame([range(1, len(sns) + 1)] * len(com_i), com_i, sns).T
-            mask = (ref <= must_stay_up) & initially_up
-            name = f"{c}-com-status-min_up_time_must_stay_up"
-            n.model.add_constraints(status, "=", 1, name, mask=mask)
+    # upper dispatch level limit
+    lhs = (1, p), (-upper_p, status)
+    n.model.add_constraints(lhs, "<=", 0, f"{c}-com-p-upper", mask=active)
 
-        min_up_time = min_up_time.clip(upper=len(sns))
-        lhs = []
-        coords = com_i[min_up_time >= 1]
-        for asset in coords:
-            up_time = min_up_time[asset]
-            # reverse snapshot order to correctly apply rolling_sum, and unreverse
-            asset_sel = {com_i.name: asset}
-            asset_status = status.sel(asset_sel, drop=True)
-            kwargs = dict(snapshot=up_time, center=False)
-            expr = asset_status.loc[::-1].rolling_sum(**kwargs).reindex(snapshot=sns)
-            # shift last var to the followed substraction
-            expr = expr.drop_isel(_term=-1)
-            lhs.append(expr - (up_time - 1) * status_diff.sel(asset_sel, drop=True))
-        lhs = merge(lhs, dim=coords).reindex({com_i.name: com_i})
+    # state-transition constraint
+    rhs = pd.DataFrame(0, sns, com_i)
+    rhs.loc[sns[0], initially_up] = -1
+    lhs = start_up - status_diff
+    n.model.add_constraints(lhs, ">=", rhs, f"{c}-com-transition-start-up", mask=active)
 
-        # rhs has to consider initial value and end-of-horizon relaxation
-        rhs = pd.DataFrame(0, sns, com_i)
-        rhs.loc[sns[0], initially_up] = -up_time_before[initially_up]
-        ref = range(1, len(sns) + 1)
-        until_end = pd.DataFrame([ref] * len(com_i), com_i, sns[::-1]).T
-        until_end = until_end.le(min_up_time).cumsum()[::-1]
-        rhs -= until_end.where(until_end < min_up_time, 0)
-        n.model.add_constraints(lhs, ">=", rhs, f"{c}-com-status-min_up_time")
+    rhs = pd.DataFrame(0, sns, com_i)
+    rhs.loc[sns[0], initially_up] = 1
+    lhs = shut_down + status_diff
+    n.model.add_constraints(
+        lhs, ">=", rhs, f"{c}-com-transition-shut-down", mask=active
+    )
 
-    if start_up_cost.sum():
-        start_up = n.model[f"{c}-start_up"]
-        lhs = start_up - status_diff
-        rhs = -initially_up.to_frame(sns[0]).T.astype(int).reindex(sns, fill_value=0)
-        n.model.add_constraints(lhs, ">=", rhs, f"{c}-com-start_up")
+    # min up time
+    expr = []
+    min_up_time_i = com_i[min_up_time_set.astype(bool)]
+    if not min_up_time_i.empty:
+        for g in min_up_time_i:
+            su = start_up.loc[:, g]
+            expr.append(su.rolling(snapshot=min_up_time_set[g]).sum())
+        lhs = -status.loc[:, min_up_time_i] + merge(expr, dim=com_i.name)
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 0, f"{c}-com-up-time", mask=active)
 
-    if min_down_time.sum() + shut_down_cost.sum():
-        # find out how long the generator has been down before snapshots
-        down_time_before_set = n.df(c)["down_time_before"].reindex(com_i)
-        until_start = n.pnl(c).status.iloc[:start_i][::-1].reindex(columns=com_i)
-        until_start = ~until_start.astype(bool)
-        ref = range(1, len(until_start) + 1)
-        down_time_before = until_start[until_start.cumsum().eq(ref, axis=0)].sum()
-        down_time_before = down_time_before.clip(
-            upper=min_down_time, lower=down_time_before_set
+    # min down time
+    expr = []
+    min_down_time_i = com_i[min_down_time_set.astype(bool)]
+    if not min_down_time_i.empty:
+        for g in min_down_time_i:
+            su = shut_down.loc[:, g]
+            expr.append(su.rolling(snapshot=min_down_time_set[g]).sum())
+        lhs = status.loc[:, min_down_time_i] + merge(expr, dim=com_i.name)
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 1, f"{c}-com-down-time", mask=active)
+
+    # up time before
+    timesteps = pd.DataFrame([range(1, len(sns) + 1)] * len(com_i), com_i, sns).T
+    if initially_up.any():
+        must_stay_up = (min_up_time_set - up_time_before_set).clip(lower=0)
+        mask = (must_stay_up >= timesteps) & initially_up
+        name = f"{c}-com-status-min_up_time_must_stay_up"
+        mask = mask & active if active is not None else mask
+        n.model.add_constraints(status, "=", 1, name, mask=mask)
+
+    # down time before
+    if initially_down.any():
+        must_stay_down = (min_down_time_set - down_time_before_set).clip(lower=0)
+        mask = (must_stay_down >= timesteps) & initially_down
+        name = f"{c}-com-status-min_down_time_must_stay_up"
+        mask = mask & active if active is not None else mask
+        n.model.add_constraints(status, "=", 0, name, mask=mask)
+
+    # linearized approximation because committable can partly start up and shut down
+    if n._linearized_uc:
+        # dispatch limit for partly start up/shut down for t-1
+        lhs = (
+            p.shift(snapshot=1)
+            - ramp_shut_down * status.shift(snapshot=1)
+            - (upper_p - ramp_shut_down) * (status - start_up)
         )
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 0, f"{c}-com-p-before", active)
 
-        initially_down = down_time_before.astype(bool)
-        must_stay_down = (min_down_time - down_time_before).clip(
-            lower=0, upper=len(sns)
+        # dispatch limit for partly start up/shut down for t
+        lhs = p - upper_p * status + (upper_p - ramp_start_up) * start_up
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 0, f"{c}-com-p-current", active)
+
+        # ramp up if committable is only partly active and some capacity is starting up
+        lhs = (
+            p
+            - p.shift(snapshot=1)
+            - (lower_p + ramp_up_limit) * status
+            + lower_p * status.shift(snapshot=1)
+            - (lower_p + ramp_up_limit - ramp_start_up) * start_up
         )
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 0, f"{c}-com-partly-start-up", active)
 
-    if min_down_time.sum():
-        if initially_down.any():
-            ref = pd.DataFrame([range(1, len(sns) + 1)] * len(com_i), com_i, sns).T
-            mask = (ref <= must_stay_down) & initially_down
-            name = f"{c}-com-status-min_down_time_must_stay_down"
-            n.model.add_constraints(status, "=", 0, name, mask=mask)
-
-        min_down_time = min_down_time.clip(upper=len(sns))
-        lhs = []
-        coords = com_i[min_down_time >= 1]
-        for asset in coords:
-            down_time = min_down_time[asset]
-            # reverse snapshot order to correctly apply rolling_sum, and unreverse
-            asset_sel = {com_i.name: asset}
-            asset_status = status.sel(asset_sel, drop=True)
-            kwargs = dict(snapshot=down_time, center=False)
-            expr = -asset_status.loc[::-1].rolling_sum(**kwargs).reindex(snapshot=sns)
-            # shift last var to the followed substraction
-            expr = expr.drop_isel(_term=-1)
-            lhs.append(expr + (down_time - 1) * status_diff.sel(asset_sel, drop=True))
-        lhs = merge(lhs, dim=coords).reindex({com_i.name: com_i})
-
-        # rhs has to consider initial value and end-of-horizon relaxation
-        rhs = -pd.DataFrame([min_down_time] * len(sns), sns, com_i)
-        rhs.loc[sns[0], initially_down] -= down_time_before[initially_down]
-        ref = range(1, len(sns) + 1)
-        until_end = pd.DataFrame([ref] * len(com_i), com_i, sns[::-1]).T
-        until_end = until_end.le(min_down_time).cumsum()[::-1]
-        rhs -= until_end.where(until_end < min_down_time, 0)
-        rhs = (rhs + 1).where(rhs < 0, rhs)
-        n.model.add_constraints(lhs, ">=", rhs, f"{c}-com-status-min_down_time")
-
-    if shut_down_cost.sum():
-        shut_down = n.model[f"{c}-shut_down"]
-        lhs = shut_down + status_diff
-        rhs = (
-            (~initially_down).to_frame(sns[0]).T.astype(int).reindex(sns, fill_value=0)
+        # ramp down if committable is only partly active and some capacity is shutting up
+        lhs = (
+            p.shift(snapshot=1)
+            - p
+            - ramp_shut_down * status.shift(snapshot=1)
+            + (ramp_shut_down - ramp_down_limit) * status
+            - (lower_p + ramp_down_limit - ramp_shut_down) * start_up
         )
-        n.model.add_constraints(lhs, ">=", rhs, f"{c}-com-shut_down")
+        lhs = lhs.sel(snapshot=sns[1:])
+        n.model.add_constraints(lhs, "<=", 0, f"{c}-com-partly-shut-down", active)
 
 
 def define_nominal_constraints_for_extendables(n, c, attr):
@@ -281,6 +315,8 @@ def define_ramp_limit_constraints(n, sns, c, attr):
         return
     if n.df(c)[["ramp_limit_up", "ramp_limit_down"]].isnull().all().all():
         return
+    if n.df(c)[["ramp_limit_up", "ramp_limit_down"]].eq(1).all().all():
+        return
 
     # ---------------- Check if ramping is at start of n.snapshots --------------- #
 
@@ -296,15 +332,23 @@ def define_ramp_limit_constraints(n, sns, c, attr):
         active = get_activity_mask(n, c, sns)
         rhs_start = pd.DataFrame(0, index=sns, columns=n.df(c).index)
         rhs_start.loc[sns[0]] = p_start
-        p_actual = lambda idx: reindex(p, c, idx)
-        p_previous = lambda idx: reindex(p, c, idx).shift(snapshot=1)
+
+        def p_actual(idx):
+            return reindex(p, c, idx)
+
+        def p_previous(idx):
+            return reindex(p, c, idx).shift(snapshot=1)
+
     else:
         active = get_activity_mask(n, c, sns[1:])
         rhs_start = pd.DataFrame(0, index=sns[1:], columns=n.df(c).index)
-        p_actual = lambda idx: reindex(p, c, idx).sel(snapshot=sns[1:])
-        p_previous = (
-            lambda idx: reindex(p, c, idx).shift(snapshot=1).sel(snapshot=sns[1:])
-        )
+        rhs_start.index.name = "snapshot"
+
+        def p_actual(idx):
+            return reindex(p, c, idx).sel(snapshot=sns[1:])
+
+        def p_previous(idx):
+            return reindex(p, c, idx).shift(snapshot=1).sel(snapshot=sns[1:])
 
     # ----------------------------- Fixed Generators ----------------------------- #
 
@@ -403,11 +447,15 @@ def define_ramp_limit_constraints(n, sns, c, attr):
         m.add_constraints(lhs, ">=", rhs, f"{c}-com-{attr}-ramp_limit_down", mask=mask)
 
 
-def define_nodal_balance_constraints(n, sns):
+def define_nodal_balance_constraints(
+    n, sns, transmission_losses=0, buses=None, suffix=""
+):
     """
     Defines nodal balance constraints.
     """
     m = n.model
+    if buses is None:
+        buses = n.buses.index
 
     args = [
         ["Generator", "p", "bus", 1],
@@ -427,6 +475,14 @@ def define_nodal_balance_constraints(n, sns):
             eff = get_as_dense(n, "Link", f"efficiency{i}", sns)
             args.append(["Link", "p", f"bus{i}", eff])
 
+    if transmission_losses:
+        args.extend(
+            [
+                ["Line", "loss", "bus0", -0.5],
+                ["Line", "loss", "bus1", -0.5],
+            ]
+        )
+
     exprs = []
 
     for arg in args:
@@ -440,32 +496,44 @@ def define_nodal_balance_constraints(n, sns):
             sign = sign * n.df(c).sign
 
         expr = DataArray(sign) * m[f"{c}-{attr}"]
-        buses = n.df(c)[column].rename("Bus")
+        cbuses = n.df(c)[column][lambda ds: ds.isin(buses)].rename("Bus")
 
         #  drop non-existent multiport buses which are ''
         if column in ["bus" + i for i in additional_linkports(n)]:
-            buses = buses[buses != ""]
-            expr = expr.sel({c: buses.index})
+            cbuses = cbuses[cbuses != ""]
+
+        expr = expr.sel({c: cbuses.index})
 
         if expr.size:
-            expr = expr.groupby_sum(buses.to_xarray())
+            # do a fast version of `expr.groupby(cbuses.to_xarray()).sum()`
+            idx = pd.MultiIndex.from_arrays(
+                [cbuses, cbuses.groupby(cbuses).cumcount()], names=["Bus", "_term"]
+            )
+            ds = (
+                expr.data.squeeze()
+                .assign_coords({c: idx})
+                .unstack(c)
+                .reset_index("_term")
+            )
+            expr = LinearExpression(ds, m)
+
             exprs.append(expr)
 
-    lhs = merge(exprs).reindex(
-        Bus=n.buses.index, fill_value=LinearExpression.fill_value
+    lhs = merge(exprs, join="outer").reindex(
+        Bus=buses, fill_value=LinearExpression.fill_value
     )
     rhs = (
         (-get_as_dense(n, "Load", "p_set", sns) * n.loads.sign)
         .groupby(n.loads.bus, axis=1)
         .sum()
-        .reindex(columns=n.buses.index, fill_value=0)
+        .reindex(columns=buses, fill_value=0)
     )
     # the name for multi-index is getting lost by groupby before pandas 1.4.0
     # TODO remove once we bump the required pandas version to >= 1.4.0
     rhs.index.name = "snapshot"
-    rhs = DataArray(rhs)
 
     empty_nodal_balance = (lhs.vars == -1).all("_term")
+    rhs = DataArray(rhs)
     if empty_nodal_balance.any():
         if (empty_nodal_balance & (rhs != 0)).any().item():
             raise ValueError("Empty LHS with non-zero RHS in nodal balance constraint.")
@@ -474,7 +542,10 @@ def define_nodal_balance_constraints(n, sns):
     else:
         mask = None
 
-    n.model.add_constraints(lhs, "=", rhs, "Bus-nodal_balance", mask=mask)
+    if suffix:
+        lhs = lhs.rename(Bus=f"Bus{suffix}")
+        rhs = rhs.rename(Bus=f"Bus{suffix}")
+    n.model.add_constraints(lhs, "=", rhs, f"Bus{suffix}-nodal_balance", mask=mask)
 
 
 def define_kirchhoff_voltage_constraints(n, sns):
@@ -497,7 +568,7 @@ def define_kirchhoff_voltage_constraints(n, sns):
     periods = sns.unique("period") if n._multi_invest else [None]
 
     for period in periods:
-        n.determine_network_topology(investment_period=period)
+        n.determine_network_topology(investment_period=period, skip_isolated_buses=True)
 
         snapshots = sns if period is None else sns[sns.get_loc(period)]
 
@@ -524,11 +595,11 @@ def define_kirchhoff_voltage_constraints(n, sns):
                     coords={"snapshot": snapshots},
                 )
                 ds = Dataset({"coeffs": coeffs, "vars": vars})
-                exprs.append(LinearExpression(ds))
+                exprs.append(LinearExpression(ds, m))
 
         if len(exprs):
             exprs = merge(exprs, dim="cycles")
-            exprs = exprs.assign_coords(cycles=range(len(exprs.cycles)))
+            exprs = exprs.assign_coords(cycles=range(len(exprs.data.cycles)))
             lhs.append(exprs)
 
     if len(lhs):
@@ -549,7 +620,6 @@ def define_fixed_nominal_constraints(n, c, attr):
     attr : str
         name of the attribute, e.g. 'p'
     """
-
     if attr + "_set" not in n.df(c):
         return
 
@@ -577,7 +647,6 @@ def define_fixed_operation_constraints(n, sns, c, attr):
     attr : str
         name of the attribute, e.g. 'p'
     """
-
     if attr + "_set" not in n.pnl(c):
         return
 
@@ -606,7 +675,6 @@ def define_storage_unit_constraints(n, sns):
 
     previous_soc + p_store - p_dispatch + inflow - spill == soc
     """
-
     m = n.model
     c = "StorageUnit"
     dim = "snapshot"
@@ -619,9 +687,9 @@ def define_storage_unit_constraints(n, sns):
     # elapsed hours
     eh = expand_series(n.snapshot_weightings.stores[sns], assets.index)
     # efficiencies
-    eff_stand = expand_series(1 - assets.standing_loss, sns).T.pow(eh)
-    eff_dispatch = expand_series(assets.efficiency_dispatch, sns).T
-    eff_store = expand_series(assets.efficiency_store, sns).T
+    eff_stand = (1 - get_as_dense(n, c, "standing_loss", sns)).pow(eh)
+    eff_dispatch = get_as_dense(n, c, "efficiency_dispatch", sns)
+    eff_store = get_as_dense(n, c, "efficiency_store", sns)
 
     soc = m[f"{c}-state_of_charge"]
 
@@ -639,9 +707,13 @@ def define_storage_unit_constraints(n, sns):
     noncyclic_b = ~assets.cyclic_state_of_charge.to_xarray()
     include_previous_soc = (active.cumsum(dim) != 1).where(noncyclic_b, True)
 
-    kwargs = dict(snapshot=1, roll_coords=False)
-    previous_soc = soc.where(active, nan).ffill(dim).roll(**kwargs).ffill(dim)
-    previous_soc = previous_soc.sanitize().where(include_previous_soc)
+    previous_soc = (
+        soc.where(active)
+        .ffill(dim)
+        .roll(snapshot=1)
+        .ffill(dim)
+        .where(include_previous_soc)
+    )
 
     # We add inflow and initial soc for noncyclic assets to rhs
     soc_init = assets.state_of_charge_initial.to_xarray()
@@ -650,25 +722,29 @@ def define_storage_unit_constraints(n, sns):
     if isinstance(sns, pd.MultiIndex):
         # If multi-horizon optimizing, we update the previous_soc and the rhs
         # for all assets which are cyclid/non-cyclid per period.
-        periods = soc.period
+        periods = soc.coords["period"]
         per_period = (
             assets.cyclic_state_of_charge_per_period.to_xarray()
             | assets.state_of_charge_initial_per_period.to_xarray()
         )
 
         # We calculate the previous soc per period while cycling within a period
-        kwargs = dict(shortcut=True, shift=1, axis=list(soc.dims).index(dim))
-        previous_soc_pp = soc.groupby(periods).map(roll, **kwargs)
+        # Normally, we should use groupby, but is broken for multi-index
+        # see https://github.com/pydata/xarray/issues/6836
+        ps = n.investment_periods.rename("period")
+        sl = slice(None, None)
+        previous_soc_pp = [soc.data.sel(snapshot=(p, sl)).roll(snapshot=1) for p in ps]
+        previous_soc_pp = concat(previous_soc_pp, dim="snapshot")
 
         # We create a mask `include_previous_soc_pp` which excludes the first
         # snapshot of each period for non-cyclic assets.
-        kwargs = dict(shortcut=True, axis=list(active.dims).index(dim))
-        include_previous_soc_pp = active.groupby(periods).map(cumsum, **kwargs) != 1
+        include_previous_soc_pp = active & (periods == periods.shift(snapshot=1))
         include_previous_soc_pp = include_previous_soc_pp.where(noncyclic_b, True)
-        previous_soc_pp = previous_soc_pp.where(include_previous_soc_pp, -1)
+        # We take values still to handle internal xarray multi-index difficulties
+        previous_soc_pp = previous_soc_pp.where(include_previous_soc_pp.values, -1)
 
         # update the previous_soc variables and right hand side
-        previous_soc = previous_soc_pp.where(per_period, previous_soc)
+        previous_soc = previous_soc.where(~per_period, previous_soc_pp.values)
         include_previous_soc = include_previous_soc_pp.where(
             per_period, include_previous_soc
         )
@@ -684,7 +760,6 @@ def define_store_constraints(n, sns):
 
     previous_e - p == e
     """
-
     m = n.model
     c = "Store"
     dim = "snapshot"
@@ -697,7 +772,7 @@ def define_store_constraints(n, sns):
     # elapsed hours
     eh = expand_series(n.snapshot_weightings.stores[sns], assets.index)
     # efficiencies
-    eff_stand = expand_series(1 - assets.standing_loss, sns).T.pow(eh)
+    eff_stand = (1 - get_as_dense(n, c, "standing_loss", sns)).pow(eh)
 
     e = m[c + "-e"]
     p = m[c + "-p"]
@@ -709,9 +784,9 @@ def define_store_constraints(n, sns):
     noncyclic_b = ~assets.e_cyclic.to_xarray()
     include_previous_e = (active.cumsum(dim) != 1).where(noncyclic_b, True)
 
-    kwargs = dict(snapshot=1, roll_coords=False)
-    previous_e = e.where(active).ffill(dim).roll(**kwargs).ffill(dim)
-    previous_e = previous_e.sanitize().where(include_previous_e)
+    previous_e = (
+        e.where(active).ffill(dim).roll(snapshot=1).ffill(dim).where(include_previous_e)
+    )
 
     # We add inflow and initial e for for noncyclic assets to rhs
     e_init = assets.e_initial.to_xarray()
@@ -719,28 +794,69 @@ def define_store_constraints(n, sns):
     if isinstance(sns, pd.MultiIndex):
         # If multi-horizon optimizing, we update the previous_e and the rhs
         # for all assets which are cyclid/non-cyclid per period.
-        periods = e.period
+        periods = e.coords["period"]
         per_period = (
             assets.e_cyclic_per_period.to_xarray()
             | assets.e_initial_per_period.to_xarray()
         )
 
         # We calculate the previous e per period while cycling within a period
-        kwargs = dict(shortcut=True, shift=1, axis=list(e.dims).index(dim))
-        previous_e_pp = e.groupby(periods).map(roll, **kwargs)
+        # Normally, we should use groupby, but is broken for multi-index
+        # see https://github.com/pydata/xarray/issues/6836
+        ps = n.investment_periods.rename("period")
+        sl = slice(None, None)
+        previous_e_pp = [e.data.sel(snapshot=(p, sl)).roll(snapshot=1) for p in ps]
+        previous_e_pp = concat(previous_e_pp, dim="snapshot")
 
         # We create a mask `include_previous_e_pp` which excludes the first
         # snapshot of each period for non-cyclic assets.
-        kwargs = dict(shortcut=True, axis=list(active.dims).index(dim))
-        include_previous_e_pp = active.groupby(periods).map(cumsum, **kwargs) != 1
+        include_previous_e_pp = active & (periods == periods.shift(snapshot=1))
         include_previous_e_pp = include_previous_e_pp.where(noncyclic_b, True)
-        previous_e_pp = previous_e_pp.where(include_previous_e_pp, -1)
+        # We take values still to handle internal xarray multi-index difficulties
+        previous_e_pp = previous_e_pp.where(include_previous_e_pp.values, -1)
 
         # update the previous_e variables and right hand side
-        previous_e = previous_e_pp.where(per_period, previous_e)
+        previous_e = previous_e.where(~per_period, previous_e_pp.values)
         include_previous_e = include_previous_e_pp.where(per_period, include_previous_e)
 
     lhs += [(eff_stand, previous_e)]
     rhs = -e_init.where(~include_previous_e, 0)
 
     m.add_constraints(lhs, "=", rhs, f"{c}-energy-balance", mask=active)
+
+
+def define_loss_constraints(n, sns, c, transmission_losses):
+    tangents = transmission_losses
+    active = get_activity_mask(n, c, sns) if n._multi_invest else None
+
+    s_max_pu = get_as_dense(n, c, "s_max_pu").loc[sns]
+
+    s_nom_max = n.df(c)["s_nom_max"].where(
+        n.df(c)["s_nom_extendable"], n.df(c)["s_nom"]
+    )
+
+    assert isfinite(
+        s_nom_max
+    ).all(), f"Loss approximation requires finite 's_nom_max' for extendable branches:\n {s_nom_max[~isfinite(s_nom_max)]}"
+
+    r_pu_eff = n.df(c)["r_pu_eff"]
+
+    upper_limit = r_pu_eff * (s_max_pu * s_nom_max) ** 2
+
+    loss = n.model[f"{c}-loss"]
+    flow = n.model[f"{c}-s"]
+
+    n.model.add_constraints(loss <= upper_limit, name=f"{c}-loss_upper", mask=active)
+
+    for k in range(1, tangents + 1):
+        p_k = k / tangents * s_max_pu * s_nom_max
+        loss_k = r_pu_eff * p_k**2
+        slope_k = 2 * r_pu_eff * p_k
+        offset_k = loss_k - slope_k * p_k
+
+        for sign in [-1, 1]:
+            lhs = n.model.linexpr((1, loss), (sign * slope_k, flow))
+
+            n.model.add_constraints(
+                lhs >= offset_k, name=f"{c}-loss_tangents-{k}-{sign}", mask=active
+            )
