@@ -5,6 +5,7 @@ Build optimisation problems from PyPSA networks with Linopy.
 """
 import logging
 import os
+from functools import wraps
 
 import numpy as np
 import pandas as pd
@@ -17,11 +18,12 @@ from pypsa.optimization.abstract import (
     optimize_security_constrained,
     optimize_transmission_expansion_iteratively,
 )
-from pypsa.optimization.common import set_from_frame
+from pypsa.optimization.common import get_strongly_meshed_buses, set_from_frame
 from pypsa.optimization.constraints import (
     define_fixed_nominal_constraints,
     define_fixed_operation_constraints,
     define_kirchhoff_voltage_constraints,
+    define_loss_constraints,
     define_nodal_balance_constraints,
     define_nominal_constraints_for_extendables,
     define_operational_constraints_for_committables,
@@ -34,11 +36,14 @@ from pypsa.optimization.constraints import (
 from pypsa.optimization.global_constraints import (
     define_growth_limit,
     define_nominal_constraints_per_bus_carrier,
+    define_operational_limit,
     define_primary_energy_limit,
+    define_tech_capacity_expansion_limit,
     define_transmission_expansion_cost_limit,
     define_transmission_volume_expansion_limit,
 )
 from pypsa.optimization.variables import (
+    define_loss_variables,
     define_nominal_variables,
     define_operational_variables,
     define_shut_down_variables,
@@ -104,7 +109,7 @@ def define_objective(n, sns):
     for c, attr in lookup.query("marginal_cost").index:
         cost = (
             get_as_dense(n, c, "marginal_cost", sns)
-            .loc[:, lambda ds: (ds != 0).all()]
+            .loc[:, lambda ds: (ds != 0).any()]
             .mul(weighting, axis=0)
         )
         if cost.empty:
@@ -135,7 +140,6 @@ def define_objective(n, sns):
     # unit commitment
     keys = ["start_up", "shut_down"]
     for c, attr in lookup.query("variable in @keys").index:
-
         com_i = n.get_committable_i(c)
         cost = n.df(c)[attr + "_cost"].reindex(com_i)
 
@@ -152,7 +156,14 @@ def define_objective(n, sns):
     m.objective = merge(objective)
 
 
-def create_model(n, snapshots=None, multi_investment_periods=False, **kwargs):
+def create_model(
+    n,
+    snapshots=None,
+    multi_investment_periods=False,
+    transmission_losses=0,
+    linearized_unit_commitment=False,
+    **kwargs,
+):
     """
     Create a linopy.Model instance from a pypsa network.
 
@@ -165,8 +176,11 @@ def create_model(n, snapshots=None, multi_investment_periods=False, **kwargs):
         A list of snapshots to optimise, must be a subset of
         network.snapshots, defaults to network.snapshots
     multi_investment_periods : bool, default False
-        Whether to optimise as a single investment period or to optimise in multiple
+        Whether to optimise as a single investment period or to optimize in multiple
         investment periods. Then, snapshots should be a ``pd.MultiIndex``.
+    transmission_losses : int, default 0
+    linearized_unit_commitment : bool, default False
+        Whether to optimise using the linearised unit commitment formulation or not.
     **kwargs:
         Keyword arguments used by `linopy.Model()`, such as `solver_dir` or `chunk`.
 
@@ -175,6 +189,7 @@ def create_model(n, snapshots=None, multi_investment_periods=False, **kwargs):
     linopy.model
     """
     sns = _as_snapshots(n, snapshots)
+    n._linearized_uc = int(linearized_unit_commitment)
     n._multi_invest = int(multi_investment_periods)
     n.consistency_check()
 
@@ -195,27 +210,60 @@ def create_model(n, snapshots=None, multi_investment_periods=False, **kwargs):
     define_spillage_variables(n, sns)
     define_operational_variables(n, sns, "Store", "p")
 
+    if transmission_losses:
+        for c in n.passive_branch_components:
+            define_loss_variables(n, sns, c)
+
     # Define constraints
     for c, attr in lookup.query("nominal").index:
         define_nominal_constraints_for_extendables(n, c, attr)
         define_fixed_nominal_constraints(n, c, attr)
 
     for c, attr in lookup.query("not nominal and not handle_separately").index:
-        define_operational_constraints_for_non_extendables(n, sns, c, attr)
-        define_operational_constraints_for_extendables(n, sns, c, attr)
+        define_operational_constraints_for_non_extendables(
+            n, sns, c, attr, transmission_losses
+        )
+        define_operational_constraints_for_extendables(
+            n, sns, c, attr, transmission_losses
+        )
         define_operational_constraints_for_committables(n, sns, c)
         define_ramp_limit_constraints(n, sns, c, attr)
         define_fixed_operation_constraints(n, sns, c, attr)
 
-    define_nodal_balance_constraints(n, sns)
+    meshed_buses = get_strongly_meshed_buses(n)
+    weakly_meshed_buses = n.buses.index.difference(meshed_buses)
+    if not meshed_buses.empty and not weakly_meshed_buses.empty:
+        # Write constraint for buses many terms and for buses with a few terms
+        # separately. This reduces memory usage for large networks.
+        define_nodal_balance_constraints(
+            n, sns, transmission_losses=transmission_losses, buses=weakly_meshed_buses
+        )
+        define_nodal_balance_constraints(
+            n,
+            sns,
+            transmission_losses=transmission_losses,
+            buses=meshed_buses,
+            suffix="-meshed",
+        )
+    else:
+        define_nodal_balance_constraints(
+            n, sns, transmission_losses=transmission_losses
+        )
+
     define_kirchhoff_voltage_constraints(n, sns)
     define_storage_unit_constraints(n, sns)
     define_store_constraints(n, sns)
 
+    if transmission_losses:
+        for c in n.passive_branch_components:
+            define_loss_constraints(n, sns, c, transmission_losses)
+
     # Define global constraints
     define_primary_energy_limit(n, sns)
     define_transmission_expansion_cost_limit(n, sns)
     define_transmission_volume_expansion_limit(n, sns)
+    define_tech_capacity_expansion_limit(n, sns)
+    define_operational_limit(n, sns)
     define_nominal_constraints_per_bus_carrier(n, sns)
     define_growth_limit(n, sns)
 
@@ -229,10 +277,9 @@ def assign_solution(n):
     Map solution to network components.
     """
     m = n.model
-    sns = n.model.parameters.snapshots
+    sns = n.model.parameters.snapshots.to_index()
 
     for name, sol in m.solution.items():
-
         if name == "objective_constant":
             continue
 
@@ -240,7 +287,6 @@ def assign_solution(n):
         df = sol.to_pandas()
 
         if "snapshot" in sol.dims:
-
             if c in n.passive_branch_components and attr == "s":
                 set_from_frame(n, c, "p0", df)
                 set_from_frame(n, c, "p1", -df)
@@ -262,9 +308,10 @@ def assign_solution(n):
             n.df(c)[attr + "_opt"].update(df)
 
     # if nominal capacity was no variable set optimal value to nominal
-    for (c, attr) in lookup.query("nominal").index:
-        if f"{c}-{attr}" not in m.variables:
-            n.df(c)[attr + "_opt"] = n.df(c)[attr]
+    for c, attr in lookup.query("nominal").index:
+        fix_i = n.get_non_extendable_i(c)
+        if not fix_i.empty:
+            n.df(c).loc[fix_i, f"{attr}_opt"] = n.df(c).loc[fix_i, attr]
 
     # recalculate storageunit net dispatch
     if not n.df("StorageUnit").empty:
@@ -282,14 +329,12 @@ def assign_duals(n):
     unassigned = []
 
     for name, dual in m.dual.items():
-
         try:
             c, attr = name.split("-", 1)
         except ValueError:
             continue
 
         if "snapshot" in dual.dims:
-
             try:
                 df = dual.transpose("snapshot", ...).to_pandas()
                 spec = attr.rsplit("-", 1)[-1]
@@ -306,7 +351,7 @@ def assign_duals(n):
 
                 if spec in assign:
                     set_from_frame(n, c, "mu_" + spec, df)
-                elif attr == "nodal_balance":
+                elif attr.endswith("nodal_balance"):
                     set_from_frame(n, c, "marginal_price", df)
             except:
                 unassigned.append(name)
@@ -356,7 +401,9 @@ def post_processing(n):
     for i in additional_linkports(n):
         ca.append(("Link", f"p{i}", f"bus{i}"))
 
-    sign = lambda c: n.df(c).sign if "sign" in n.df(c) else -1  # sign for 'Link'
+    def sign(c):
+        return n.df(c).sign if "sign" in n.df(c) else -1  # sign for 'Link'
+
     n.buses_t.p = (
         pd.concat(
             [
@@ -392,8 +439,12 @@ def optimize(
     n,
     snapshots=None,
     multi_investment_periods=False,
+    transmission_losses=0,
+    linearized_unit_commitment=False,
     model_kwargs={},
     extra_functionality=None,
+    solver_name="glpk",
+    solver_options={},
     **kwargs,
 ):
     """
@@ -408,6 +459,13 @@ def optimize(
     multi_investment_periods : bool, default False
         Whether to optimise as a single investment period or to optimise in multiple
         investment periods. Then, snapshots should be a ``pd.MultiIndex``.
+    transmission_losses : int, default 0
+        Whether an approximation of transmission losses should be included
+        in the linearised power flow formulation. A passed number will denote
+        the number of tangents used for the piecewise linear approximation.
+        Defaults to 0, which ignores losses.
+    linearized_unit_commitment : bool, default False
+        Whether to optimise using the linearised unit commitment formulation or not.
     model_kwargs: dict
         Keyword arguments used by `linopy.Model`, such as `solver_dir` or `chunk`.
     extra_functionality : callable
@@ -416,6 +474,10 @@ def optimize(
         the model building is complete, but before it is sent to the
         solver. It allows the user to
         add/change constraints and add/change the objective function.
+    solver_name : str
+        Name of the solver to use.
+    solver_options : dict
+        Keyword arguments used by the solver. Can also be passed via **kwargs.
     **kwargs:
         Keyword argument used by `linopy.Model.solve`, such as `solver_name`,
         `problem_fn` or solver options directly passed to the solver.
@@ -427,13 +489,20 @@ def optimize(
 
     sns = _as_snapshots(n, snapshots)
     n._multi_invest = int(multi_investment_periods)
+    n._linearized_uc = linearized_unit_commitment
 
     n.consistency_check()
-    m = create_model(n, sns, multi_investment_periods, **model_kwargs)
+    m = create_model(
+        n,
+        sns,
+        multi_investment_periods,
+        transmission_losses,
+        linearized_unit_commitment,
+        **model_kwargs,
+    )
     if extra_functionality:
         extra_functionality(n, sns)
-    kwargs.setdefault("solver_name", "glpk")
-    status, condition = m.solve(**kwargs)
+    status, condition = m.solve(solver_name=solver_name, **solver_options, **kwargs)
 
     if status == "ok":
         assign_solution(n)
@@ -443,14 +512,6 @@ def optimize(
     return status, condition
 
 
-def is_documented_by(original):
-    def wrapper(target):
-        target.__doc__ = original.__doc__
-        return target
-
-    return wrapper
-
-
 class OptimizationAccessor:
     """
     Optimization accessor for building and solving models using linopy.
@@ -459,29 +520,31 @@ class OptimizationAccessor:
     def __init__(self, network):
         self._parent = network
 
+    @wraps(optimize)
     def __call__(self, *args, **kwargs):
         return optimize(self._parent, *args, **kwargs)
 
-    __call__.__doc__ = optimize.__doc__
+    @wraps(create_model)
+    def create_model(self, *args, **kwargs):
+        return create_model(self._parent, *args, **kwargs)
 
-    @is_documented_by(create_model)
-    def create_model(self, **kwargs):
-        return create_model(self._parent, **kwargs)
-
-    def solve_model(self, **kwargs):
+    def solve_model(self, solver_name="glpk", solver_options={}, **kwargs):
         """
         Solve an already created model and assign its solution to the network.
 
         Parameters
         ----------
+        solver_name : str
+            Name of the solver to use.
+        solver_options : dict
+            Keyword arguments used by the solver. Can also be passed via **kwargs.
         **kwargs:
             Keyword argument used by `linopy.Model.solve`, such as `solver_name`,
             `problem_fn` or solver options directly passed to the solver.
         """
         n = self._parent
         m = n.model
-        kwargs.setdefault("solver_name", "glpk")
-        status, condition = m.solve(**kwargs)
+        status, condition = m.solve(solver_name=solver_name, **solver_options, **kwargs)
 
         if status == "ok":
             assign_solution(n)
@@ -490,23 +553,23 @@ class OptimizationAccessor:
 
         return status, condition
 
-    @is_documented_by(assign_solution)
+    @wraps(assign_solution)
     def assign_solution(self, **kwargs):
         return assign_solution(self._parent, **kwargs)
 
-    @is_documented_by(assign_duals)
+    @wraps(assign_duals)
     def assign_duals(self, **kwargs):
         return assign_duals(self._parent, **kwargs)
 
-    @is_documented_by(post_processing)
+    @wraps(post_processing)
     def post_processing(self, **kwargs):
         return post_processing(self._parent, **kwargs)
 
-    @is_documented_by(optimize_transmission_expansion_iteratively)
+    @wraps(optimize_transmission_expansion_iteratively)
     def optimize_transmission_expansion_iteratively(self, *args, **kwargs):
         optimize_transmission_expansion_iteratively(self._parent, *args, **kwargs)
 
-    @is_documented_by(optimize_security_constrained)
+    @wraps(optimize_security_constrained)
     def optimize_security_constrained(self, *args, **kwargs):
         optimize_security_constrained(self._parent, *args, **kwargs)
 
@@ -577,7 +640,7 @@ class OptimizationAccessor:
             suffix,
             bus=buses,
             carrier="load",
-            sig=sign,
+            sign=sign,
             marginal_cost=marginal_cost,
             p_nom=p_nom,
         )
