@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-
 """
 Statistics Accessor.
 """
@@ -9,7 +8,7 @@ __author__ = (
     "PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html"
 )
 __copyright__ = (
-    "Copyright 2015-2022 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
+    "Copyright 2015-2023 PyPSA Developers, see https://pypsa.readthedocs.io/en/latest/developers.html, "
     "MIT License"
 )
 
@@ -268,7 +267,7 @@ class StatisticsAccessor:
     def opex(
         self,
         comps=None,
-        aggregate_time="mean",
+        aggregate_time="sum",
         aggregate_groups="sum",
         groupby=None,
     ):
@@ -299,7 +298,7 @@ class StatisticsAccessor:
     def supply(
         self,
         comps=None,
-        aggregate_time="mean",
+        aggregate_time="sum",
         aggregate_groups="sum",
         groupby=None,
     ):
@@ -330,7 +329,7 @@ class StatisticsAccessor:
     def withdrawal(
         self,
         comps=None,
-        aggregate_time="mean",
+        aggregate_time="sum",
         aggregate_groups="sum",
         groupby=None,
     ):
@@ -361,15 +360,15 @@ class StatisticsAccessor:
     def curtailment(
         self,
         comps=None,
-        aggregate_time="mean",
+        aggregate_time="sum",
         aggregate_groups="sum",
         groupby=None,
     ):
         """
         Calculate the curtailment of components in the network.
 
-        The calculation only considers assets with a `p_max_pu` time series,
-        which is used to quantify the available power potential.
+        The calculation only considers assets with a `p_max_pu` time
+        series, which is used to quantify the available power potential.
 
         For information on the list of arguments, see the docs in
         `Network.statistics` or `pypsa.statitics.StatisticsAccessor`.
@@ -422,7 +421,7 @@ class StatisticsAccessor:
     def revenue(
         self,
         comps=None,
-        aggregate_time="mean",
+        aggregate_time="sum",
         aggregate_groups="sum",
         groupby=None,
     ):
